import logging


def Handle(
    logger,
    handler=logging.NullHandler(),
    formatter="%(asctime)s %(name)s - %(levelname)s: %(message)s",
    level="WARNING",
):
    """
    Handle a logger with a standardised formatting.

    Parameters
    -----------
    logger : :class:`logging.Logger` | :class:`str`
        Logger or module name to source a logger from.
    handler : :class:`logging.Handler`
        Handler for the logging messages.
    formatter : :class:`str` | :class:`logging.Formatter`
        Formatter for the logging handler. Strings will be passed to
        the :class:`logging.Formatter` constructor.
    level : :class:`str`
        Logging level for the handler.

    Returns
    ----------
    :class:`logging.Logger`
        Configured logger.
    """
    if isinstance(logger, str):
        logger = logging.getLogger(logger)
    elif isinstance(logger, logging.Logger):
        pass
    else:
        raise NotImplementedError
    if isinstance(formatter, str):
        formatter = logging.Formatter(formatter)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(getattr(logging, level))
    return logger



"""
Utilities for reading alphaMELTS table outputs.
"""
import os, sys
import re
import io
import pandas as pd
from pathlib import Path
import pyrolite.geochem
from ..parse import from_melts_cstr
from ..meltsfile import df_to_meltsfiles
from pyrolite.util.pd import zero_to_nan, to_frame, to_ser

from .load import import_tables, aggregate_tables
from ..util.log import Handle

logger = Handle(__name__)



"""
Utilities for reading and writing .melts files.
"""
import io
import os
import itertools
import numpy as np
import pandas as pd
from pathlib import Path
from pyrolite.util.pd import to_frame, to_ser
from pyrolite.geochem.ind import common_elements, common_oxides
import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())
logger = logging.getLogger(__name__)


def dict_to_meltsfile(
    d, linesep=os.linesep, writetraces=True, modes=[], exclude=[], **kwargs
):
    """
    Converts a dictionary to a MELTSfile text representation. It requires 'title'
    and 'initial composition' lines, major elements to be represented as oxides
    in Wt% and trace elements in µg/g.

    Parameters
    ----------
    d : :class:`dict`
        Dictionary to convert to a melts file.
    linesep : :class:`str`
        Line separation character.
    writetraces : :class:`bool`
        Whether to include traces in the output file.
    modes : :class:`list`
        List of modes to use (e.g. 'isobaric', 'fractionate solids').
    exclude : :class:`list`
        List of chemical components to exclude from the meltsfile.

    Returns
    -------
    :class:`str`
        String representation of the meltsfile, which can be immediately written to a
        file object.

    Notes
    -------

        * Some of the parameters are one-to-many, including modes, phase fractionation,
            supression and coexist-limits.

    Todo
    -----
        * Parameter validation.
    """
    # we'll incrementally collect the lines for the melts file from the data dictionary
    lines = []
    # first we add the title
    assert ("Title" in d) or ("title" in d)
    if "Title" in d:
        lines.append("Title: {}".format(d["Title"]))
    else:
        lines.append("Title: {}".format(d["title"]))

    # then we'll collect the composition Parameters
    majors = [
        (k, v) for (k, v) in d.items() if k in common_oxides() and not k in exclude
    ]
    traces = [
        (k, v) for (k, v) in d.items() if k in common_elements() and not k in exclude
    ]
    for k, v in majors:
        if not pd.isnull(v):  # no NaN data in MELTS files
            lines.append("Initial Composition: {} {}".format(k, v))

    if writetraces:
        for k, v in traces:
            if not pd.isnull(v):  # no NaN data in MELTS files
                lines.append("Initial Trace: {} {}".format(k, v))

    # follwed by the pressure and temperature parameters
    PTpars = [
        (" ".join([pre, param]), d.get(" ".join([pre, param]), None))
        for pre, param in itertools.product(
            ["Initial", "Final", "Increment"], ["Temperature", "Pressure"]
        )
    ]

    for (k, v) in PTpars:
        if not pd.isnull(v):  # no NaN data in MELTS files
            lines.append("{}: {}".format(k, v))

    for mfilepar in [
        "dp/dt",
        "Log fO2 Path",
        "Log fO2 Delta",
        "Suppress",
        "Limit coexisting",
        "Fractionate",
    ]:
        par = [(k, v) for (k, v) in d.items() if k.lower() == mfilepar.lower()]
        if par:
            par, v = par[0]
            if isinstance(v, (list, set, tuple)):
                for iv in v:
                    if not pd.isnull(iv):  # no NaN data in MELTS files
                        lines.append(
                            "{}: {}".format(mfilepar, iv)
                        )  # suppress, fractionate
            else:
                if not pd.isnull(v):
                    lines.append("{}: {}".format(mfilepar, v))

    for m in modes:
        lines.append("Mode: {}".format(m))

    # valid_modes = ["Fractionate Solids", "Fractionate"]
    return linesep.join(lines)


def ser_to_meltsfile(
    ser, linesep=os.linesep, writetraces=True, modes=[], exclude=[], **kwargs
):
    """
    Converts a series to a MELTSfile text representation. It requires 'title'
    and 'initial composition' lines, major elements to be represented as oxides
    in Wt% and trace elements in µg/g.

    Parameters
    ----------
    ser : :class:`pandas.Series`
        Series to convert to a melts file.
    linesep : :class:`str`
        Line separation character.
    writetraces : :class:`bool`
        Whether to include traces in the output file.
    modes : :class:`list`
        List of modes to use (e.g. 'isobaric', 'fractionate solids').
    exclude : :class:`list`
        List of chemical components to exclude from the meltsfile.

    Returns
    -------
    :class:`str`
        String representation of the meltsfile, which can be immediately written to a
        file object.

    Todo
    -----
        * Parameter validation.
    """
    lines = []
    ser = to_ser(ser)
    return dict_to_meltsfile(
        ser.to_dict(),
        linesep=linesep,
        writetraces=writetraces,
        modes=modes,
        exclude=exclude,
        **kwargs
    )


def df_to_meltsfiles(df, linesep=os.linesep, **kwargs):
    """
    Creates a number of melts files from a dataframe.

    Parameters
    -----------
    df : :class:`pandas.DataFrame`
        Dataframe from which to take the rows and create melts files.
    linesep : :class:`str`
        Line separation character.

    Returns
    -------
    :class:`list`
        List of strings which can be written to file objects.
    """

    # Type checking such that series will be passed directly to MELTSfiles
    if isinstance(df, pd.DataFrame):
        return [
            ser_to_meltsfile(df.iloc[ix, :], linesep=os.linesep, **kwargs)
            for ix in range(df.index.size)
        ]
    elif isinstance(df, pd.Series):
        return [ser_to_meltsfile(df, linesep=os.linesep, **kwargs)]


def from_meltsfile(filename):
    """
    Read from a meltsfile into a :class:`pandas.DataFrame`.

    Parameters
    -----------
    filename : :class:`str` | :class:`pathlib.Path` | :class:`io.BytesIO`
        Filename, filepath or bytes object to read from.

    Returns
    --------
    :class:`pandas.DataFrame`
        Dataframe containing meltsfile parameters.
    """
    if isinstance(filename, io.BytesIO):
        file = filename.getvalue().decode()
    elif isinstance(filename, io.StringIO):
        file = filename.getvalue()
    else:
        try:  # filepath
            with open(filename) as fh:
                file = fh.read()
        except FileNotFoundError:  # string specification of meltsfile
            file = filename

    lines = [line.split(": ") for line in file.splitlines() if line.strip()]
    fmtlines = []
    for ix, args in enumerate(lines):
        if args[0].strip().lower() in ["initial composition", "initial trace"]:
            fmtlines.append(args[1].strip().split())
        else:
            fmtlines.append([i.strip() for i in args])
    df = (
        pd.DataFrame.from_records(fmtlines)
        .set_index(0, drop=True)
        .apply(pd.to_numeric, errors="ignore")[1]
    )
    return df



from pyrolite.util.text import normalise_whitespace, to_width
from pathlib import Path


def mvar(**kwargs):
    """A data dictonary specifying MELTS environment variables."""
    d = dict(
        validate=None,
        desc=None,
        type=None,
        units=None,
        set=True,
        default=None,
        overridden_by=None,
        dependent_on=None,
        priorty=0,
    )
    d.update(kwargs)
    return d


def description(text, width=69):
    return to_width(normalise_whitespace(text), width=width)


def filepath_validator(path):
    """Validator for file paths."""
    if isinstance(path, str):
        path = Path(path)
    return path.exists() and path.isfile()


def dirpath_validator(path):
    """Validator for directory paths."""
    if isinstance(path, str):
        path = Path(path)
    return path.exists() and path.isdir()


def numeric_validator(value):
    """Validator for numeric values."""
    return isinstance(float(value), float) or isinstance(int(value), int)


def typecast(type, item):
    """Typecasts argument input for lambda functions."""

    if isinstance(item, list):
        # Return each of the function items as Typecast
        # equivalents
        return [lambda x: f(type(x)) if callable(f) else f for f in item]
    else:
        return lambda x: item(type(x))


## The following is directly form the alphaMELTS Documentation v1.8
## Written by Paula Antoshechkina and Paul Asimow October 13, 2017
MELTS_environment_variables = {  # Solution Models
    "VERSION": mvar(
        default="pMELTS",
        validate=lambda x: x in ["MELTS", "pMELTS"],
        desc=description(
            """Set to choose the liquid thermodynamic model.
           pMELTS is only recommended for peridotite bulk
           compositions between 1 and 4 GPa. If doing pHMELTS
           calculations then use ‘pMELTS’ for melting and
           ‘MELTS’ for low-pressure crystallization."""
        ),
    ),
    "OLD_GARNET": mvar(
        set=False,
        type=bool,
        desc=description(
            """Uses the old, incorrect, garnet model. When set,
              the behavior will resemble that of all GUI
              versions of MELTS released before 2005 (including
              Java MELTS) and the default behavior of
              Adiabat_1ph versions 1.4 – 2.X. When unset,
              will behave as if ADIABAT_FIX_GARNET was set in
              older (pre-3.0) versions of Adiabat_1ph."""
        ),
    ),
    "OLD_SPINEL": mvar(
        set=False,
        type=bool,
        desc=description(
            """Use the spinel model from Sack and
              Ghiorso [1991], with perfectly ideal
              mixing of volumes. When set, gives the
              same behavior as Java MELTS and
              earlier (pre-3.0) versions of
              Adiabat_1ph. When unset, uses the
              spinel volume model as formulated in
              Ghiorso and Sack [1991]."""
        ),
    ),
    "OLD_BIOTITE": mvar(
        set=False,
        type=bool,
        desc=description(
            """Uses the biotite model from Sack and Ghiorso
           [1989] instead of the default alphaMELTS one.
           When set, the program uses the same code as
           GUI versions of MELTS (2005-present). The
           behavior may differ slightly from earlier GUI
           versions of MELTS as a small error was fixed."""
        ),
    ),
    "2_AMPH": mvar(
        set=False,
        type=bool,
        desc=description(
            """Uses separate orthoamphibole and clinoamphibole
              phases. When set, gives the same behavior as Java
              MELTS and Corba MELTS applets, and earlier versions
              of Adiabat_1ph (which inherited its code from the
              Java MELTS branch). When unset, will use a single
              cummingtonite-grunerite-tremolite solution and deduce
              the amphibole structure from the energetics."""
        ),
    ),
    "NO_CHLORITE": mvar(
        set=False,
        type=bool,
        desc=description(
            """As of version 1.6, the chlorite model of
               Hunziker [2003] is included in alphaMELTS
               (see Smith et al. [2007]). To reproduce previous
               results for bulk compositions that may stabilize
               chlorite, add a ‘Suppress: chlorite’ line to the
               melts_file or set this variable."""
        ),
    ),
    # P-T Path and fO2 control
    "MODE": mvar(
        validate=lambda x: x
        in [
            "geothermal",
            "isenthalpic",
            "isentropic",
            "isobaric",
            "isochoric",
            "isothermal",
            "ptfile",
            "ptgrid",
            "tpgrid",
        ],
        default="isentropic",
        desc=description(
            """Sets the calculation mode. This variable is case
             insensitive. ‘PTfile’ (‘PTPath’ is also accepted
             for backwards compatibility) reads in P and T from
             a file; see below. ‘PTgrid’ and ‘TPgrid’ perform
             calculations on a grid bounded by the starting T
             and ALPHAMELTS_MINT/MAXT, and by the starting P and
             ALPHAMELTS_MINP/MAXP. For ‘PTgrid’ the T-loop is
             inside the P-loop, whereas ‘TPgrid’ is the other
             way round; in either case the (superliquidus or
             subsolidus) starting solution migrates with the
             outer loop. Generally, unless liquid is suppressed,
             ‘PTgrid’ with ALPHAMELTS_DELTAT < 0 will be the
             easiest combination. For other modes, there are
             various ways to initialize P, T and / or reference
             entropy, enthalpy or volume, and the thermodynamic
             path is set using ALPHAMELTS_DELTAP and / or
             ALPHAMELTS_DELTAT."""
        ),
    ),
    "PTPATH_FILE": mvar(
        set=False,
        validate=filepath_validator,
        desc=description(
            """Gives the name of the ptpath_file, which is a simple
            space delimited text file with one ‘P_value T_value’
            pair per line. If ALPHAMELTS_DELTAP and
            ALPHAMELTS_DELTAT are both zero the user will be asked
            for a maximum number of iterations to perform."""
        ),
    ),
    "DELTAP": mvar(
        validate=numeric_validator,
        type=float,
        default=1000.0,
        units="bars",
        desc=description(
            """This sets the pressure increment for isentropic,
               isothermal, geothermal or phase diagram mode.
               This is a signed number; i.e., a positive value steps
               upwards in P, negative steps down. If using a ptpath_file,
               a non-zero ALPHAMELTS_DELTAP means the whole will be read
               in and executed. Setting ALPHAMELTS_DELTAP to zero, with a
               non-zero value for ALPHAMELTS_DELTAT, means that phase
               diagram mode will step in temperature instead of pressure
               (except for liquid)."""
        ),
    ),
    "DELTAT": mvar(
        validate=numeric_validator,
        default=10.0,
        type=float,
        units="Degrees Celsius",
        desc=description(
            """This sets the temperature increment for isobaric,
               isochoric, geothermal or phase diagram mode. This is a
               signed number; i.e., a positive value steps upwards in T,
               negative steps down. If using a ptpath_file, a non-zero
               ALPHAMELTS_DELTAT means the whole will be read in and
               executed. For phase diagram mode to step in temperature
               ALPHAMELTS_DELTAP must be set to zero."""
        ),
    ),
    "MAXP": mvar(
        validate=numeric_validator,
        type=float,
        default=lambda env: [30000, 40000][env["VERSION"] != "MELTS"],
        dependent_on=["VERSION"],
        units="bars",
        desc=description(
            """Sets the maximum pressure the program will go to on
                 execution."""
        ),
    ),
    "MINP": mvar(
        validate=numeric_validator,
        type=float,
        default=lambda env: [1, 10000][env["VERSION"] != "MELTS"],
        dependent_on=["VERSION"],
        units="bars",
        desc=description(
            """Sets the minimum pressure the program will go to on
                 execution."""
        ),
    ),
    "MAXT": mvar(
        validate=numeric_validator,
        type=float,
        default=2000,
        units="Degrees Celsius",
        desc=description(
            """Sets the maximum temperature the program will go to on
                  execution."""
        ),
    ),
    "MINT": mvar(
        validate=numeric_validator,
        type=float,
        default=0,
        units="Degrees Celsius",
        desc=description(
            """Sets the minimum temperature the program will go to on
                  execution."""
        ),
    ),
    "ALTERNATIVE_FO2": mvar(
        set=False,
        type=bool,
        desc=description(
            """Normally, the parameterization of Kress and
            Carmichael [1991] is used to calculate fO2 in the
            liquid. If conditions are subsolidus or liquid is
            suppressed then the approach detailed in Asimow and
            Ghiorso [1998] is used to construct an appropriate
            redox reaction to solve for fO2 of the bulk
            assemblage. If this environment variable is set,
            however, then the method of Asimow and Ghiorso
            [1998] is used to calculate fO2 regardless of
            whether liquid is present and so, in theory, allows
            for a smoother transition across the solidus. """
        ),
    ),
    "LIQUID_FO2": mvar(
        set=False,
        type=bool,
        overridden_by=["LIQUID_FO2"],
        desc=description(
            """The method of Asimow and Ghiorso
           [1998] is computationally more involved than the
           parameterization of Kress and Carmichael [1991] and it
           is not uncommon for fO2 calculations to be successful
           with liquid present but fail subsolidus. It is possible
           to turn off the fO2 buffer manually (option 5).
           Alternatively, if this environment variable is set then
           the fO2 buffer, as formulated in Kress and Carmichael
           [1991], will only be imposed when liquid is present.
           Note that setting this variable does not change the
           fO2 buffer setting (e.g. ‘FMQ’); the program just ignores
            the flag if no liquid is around."""
        ),
    ),
    "IMPOSE_FO2": mvar(
        set=False,
        type=bool,
        desc=description(
            """Normally, for isentropic, isenthalpic and isochoric
           modes any fO2 buffer will be switched off on execution
           once the reference entropy, enthalpy or volume has been
           set (usually after the first calculation or before if set
           manually). If this environment variable is set then the
           program will alternate between (1) an unbuffered
           isenthalpic / isentropic / isochoric step and (2) an
           isobaric / isothermal fO2 buffered step. Overall this
           approximates an isenthalpic, isentropic or isochoric
           path with a desired fO2 buffer."""
        ),
    ),
    "FO2_PRESSURE_TERM": mvar(
        set=False,
        type=bool,
        desc=description(
            """In most versions of MELTS, including Corba
              MELTS and alphaMELTS, by default, reference fO2
              buffers in the system Fe-Si-O are calculated from
              the equations given in Myers and Eugster [1983].
              In the standalone GUI version of MELTS a pressure
              term is added to the fayalite-magnetite-quartz
              buffer (FMQ) to give:

              log(fO2) = − 24441.9/T + 0.110 (P − 1)/T + 8.290

              When this environment variable is set then
              alphaMELTS uses the equation as shown whereas by
              default the second term is omitted. Inclusion of
              the pressure term can have subtle effects on the
              stability of certain phases such as pyroxenes.
              Note that in Kessel et al. [2001] the coefficient
              on the pressure term is 0.05, rather than 0.110;
              the value of 0.110 was chosen by Ghiorso to
              maintain consistency with Berman [1988]."""
        ),
    ),
    # Open vs closed system
    "CONTINUOUS_MELTING": mvar(
        set=False,
        type=bool,
        desc=description(
            """By default, batch melting equations are used.
               Setting this environment variable will change the
               melting mode to continuous or fractional, where
               melt is extracted after each equilibrium. Set
               ALPHAMELTS_MINF to 0 for perfect fractional
               melting. In practice though, the program will
               run more smoothly if ALPHAMELTS_MINF is slightly
               greater than 0. See the next four environment
               variables’ entries for more details."""
        ),
    ),
    "MINF": mvar(
        default=0.005,
        type=float,
        validator=typecast(
            float, [numeric_validator, lambda x: (x >= 0.0) & (x < 1.0)]
        ),
        overridden_by=["MINPHI"],
        desc=description(
            """If ALPHAMELTS_CONTINUOUS_MELTING is set, then by default a
         fixed melt fraction, by mass, marks the threshold above which
         melt is extracted. This variable is used to change the amount
         of melt retained. If the current melt fraction is less than
         ALPHAMELTS_MINF then all the melt will be retained until the
         next step, otherwise the amount of melt removed (approximately
         F - ALPHAMELTS_MINF) will be adjusted so such that the melt
         fraction is exactly ALPHAMELTS_MINF after extraction."""
        ),
    ),
    "MINPHI": mvar(
        set=False,
        type=float,
        validator=typecast(
            float, [numeric_validator, lambda x: (x >= 0.0) & (x < 1.0)]
        ),
        overridden_by=["CONTINUOUS_RATIO"],
        desc=description(
            """If ALPHAMELTS_CONTINUOUS_MELTING is set, then set this
           environment variable controls the retained melt fraction,
           by volume i.e. the ‘residual porosity’. If the current melt
           fraction is less than ALPHAMELTS_MINPHI then all the melt
           will be retained until the next step, otherwise the amount
           of melt removed (approximately phi - ALPHAMELTS_MINPHI) will
           be adjusted so that the melt fraction, by volume, is exactly
           ALPHAMELTS_MINPHI after extraction."""
        ),
    ),
    "CONTINUOUS_RATIO": mvar(
        set=False,
        type=float,
        validator=typecast(
            float, [numeric_validator, lambda x: (x >= 0.0) & (x < 1.0)]
        ),
        overridden_by=["CONTINUOUS_VOLUME"],
        desc=description(
            """This implements another alternative definition
             of continuous melting. Instead of extracting all
             liquid above a fixed mass or volume fraction, this
             option, if set, causes the program to multiply the
             liquid mass by a fixed ratio."""
        ),
    ),
    "CONTINUOUS_VOLUME": mvar(
        set=False,
        type=float,
        desc=description(
            """"This option, if set, extracts the required
              amount of melt to retain a constant total volume.
              This reference volume is set the first time
              melting occurs and is equal to the solid volume
              plus whatever melt volume is retained according
              to the ALPHAMELTS_MINPHI variable (or a default
              value of 0.002 if that is unset). Only for
              isobaric or isothermal calculations. Note that
              this is not an ‘isochoric’ calculation as far
              as alphaMELTS is concerned because melting is
              still allowed to cause expansion; this option
              only controls how much melt must be extracted
              to return to the original volume and, if
              necessary, also adjusts pressure (for isothermal
              calculations) or temperature (for isobaric
              calculations) to maintain equilibrium."""
        ),
    ),
    "FRACTIONATE_SOLIDS": mvar(
        set=False,
        type=bool,
        desc=description(
            """To turn on fractional crystallization of all
               solid phases, set this option to true (does not
               include water, see below). Do not use this
               option if you wish to selectively fractionate
               just a few phases; instead put ‘Fractionate:
               phase’ lines in you melts_file (see the ‘MELTS
               file’ section) or adjust individual phase
               settings with menu option 8."""
        ),
    ),
    "MASSIN": mvar(
        default=0.001,
        type=float,
        validator=typecast(
            float, [numeric_validator, lambda x: (x >= 0.0) & (x < 1.0)]
        ),
        overridden_by=["MINW"],
        desc=description(
            """Set to the mass in grams of each solid phase that is
                   retained during fractional crystallization. An increased
                   value may help to stabilize the calculation. A smaller value
                   can also be used, but a minimum of 10-6 grams is recommended.
                   nce the phase is no longer in the equilibrium assemblage it
                   will be completely exhausted, regardless of the
                   ALPHAMELTS_MASSIN value."""
        ),
    ),
    "FRACTIONATE_WATER": mvar(
        set=False,
        type=bool,
        desc=description(
            """To remove free water at each calculation stage
                              in an analogous way to how melt is removed during
                              continuous melting, set this variable true. (Note
                              that, in MELTS and pMELTS, water is treated like
                              a ‘solid’, in the sense that it is not melt, so
                              you can achieve the same effect by putting a
                              ‘Fractionate: water’ line in your melts_file.
                              However, water is treated differently from the
                              other mineral phases in that it may be extracted
                              during melting or crystallization.)"""
        ),
    ),
    "MINW": mvar(
        set=False,
        type=float,
        validator=typecast(
            float, [numeric_validator, lambda x: (x >= 0.0) & (x < 1.0)]
        ),
        desc=description(
            """Set to the proportion of retained water, relative to the
                 total system mass. Works in a similar way, for water, as
                 ALPHAMELTS_MINF does for melt; may be set to exactly zero. If
                 not set then fractionation of water is treated in a similar way
                 to fractionation of a solid phase i.e. a nominal mass of 10-4
                 grams is retained at each stage to stabilize the calculation;
                 the smaller mass of water retained, compared to other solid
                 phases, reflects its significantly lower molecular mass."""
        ),
    ),
    "FRACTIONATE_TARGET": mvar(
        set=False,
        type=float,
        desc=description(
            """During normal forward fractional
                               crystallization a (negative) increment is added
                               to the temperature at each step,
                               ALPHAMELTS_DELTAT, and the run is terminated when
                               the temperature goes below ALPHAMELTS_MINT. If
                               this environment variable is set then forward or
                               backward fractionation is performed so that the
                               MgO content (in wt %) or the Mg# (in mol %) hits
                               a particular target; see the next two entries.
                               When option 3 is called alphaMELTS will perform
                               a single calculation for the current P-T
                               conditions, in the normal way. When option 4 is
                               used it will first try to find the liquidus
                               instead and use that to decide whether forward
                               or backwards fractionation is required to move
                               the liquid towards the target composition.
                               For forward fractionation, temperature will be
                               reduced, by an amount equal to the absolute value
                               of ALPHAMELTS_DELTAT, each time. For backwards
                               fractionation the program will step down (by the
                               same temperature increment) until one or more
                               ‘allowed’ solid phases join the assemblage. The
                               ‘allowed’ solid phases are any MgO-brearing
                               phases that have been ‘allowed’ by setting
                               ALPHAMELTS_FRACTIONATE_SOLIDS or by having
                               ‘Fractionate:’ lines in the melts_file. The
                               routine then assimilates these phases before
                               searching for the new liquidus. For H2O-rich
                               compositions, the liquidus temperature will be
                               for a melt composition that is water-saturated,
                               but not oversaturated enough to exsolve vapor (it
                               may also help to buffer aH2O).
                               Output will be written each time the new liquidus
                               is found. Forward- or backward-fractionated trace
                               element compostions will be calculated if
                               ALPHAMELTS_DO_TRACE is on. Execution continues
                               until the target MgO or Mg# is reached or just
                               passed; therefore, the smaller ALPHAMELTS_DELTAT
                               is, the closer the liquid composition will be to
                               the target. In the limit as the step size tends
                               to zero a liquid composition that is on, say,
                               the plagioclase + clinopyroxene cotectic would
                               be expected to stay on the cotectic as back
                               fractionation proceeds (at least until a thermal
                               divide is encountered); in practice one or other
                               solid phase may occasionally be dropped from the
                               assemblage. For complicated peritectic
                               relationships the ‘Amoeba’ routine (menu option
                               17) is more useful for constraining a plausible,
                               but still non-unique, parental melt."""
        ),
    ),
    "MGO_TARGET": mvar(
        default=8.0,
        type=float,
        validator=typecast(
            float, [lambda x: isinstance(x, float), lambda x: (x > 0.0) & (x < 100.0)]
        ),
        overridden_by=["MGNUMBER_TARGET"],
        desc=description(
            """Sets the target MgO content of the liquid for forward
                       or backward fractionation to the value in wt %. When
                       using ‘Amoeba’ (menu option 17) the value of
                       ALPHAMELTS_MGO_TARGET is used to set the MgO of the
                       parental liquid composition and the target MgO content
                       for the evolved liquid composition is taken from the
                       melts_file(s). Once ‘Amoeba’ is finished, reading in
                       another melts_file will revert to using
                       ALPHAMELTS_MGO_TARGET as the stop point for forward or
                       backward fractional crystallization."""
        ),
    ),
    "MGNUMBER_TARGET": mvar(
        set=False,
        type=float,
        validator=typecast(
            float, [lambda x: isinstance(x, float), lambda x: (x > 0.0) & (x < 100.0)]
        ),
        desc=description(
            """Sets the target Mg# of the liquid for forward or
                            backward fractionation to the value in mol %. When
                            using ‘Amoeba’ the value of
                            ALPHAMELTS_MGNUMBER_TARGET is used in a comparable
                            way to ALPHAMELTS MGO_TARGET above."""
        ),
    ),
    "ASSIMILATE": mvar(
        set=False,
        type=bool,
        desc=description(
            """This environment variable causes a user-defined mass
                       of a second bulk composition to be added after each
                       calculation stage (see ‘Bulk composition’). It is
                       intended for calculations at specified P-T conditions
                       (e.g. isobaric, isothermal or PTpath modes) or for
                       heat-balanced assimilation in isenthalpic mode. It also
                       works for isentropic or isochoric constraints under
                       certain circumstances but these options are under
                       development and, as yet, untested. Melt may be extracted
                       or solid phases fractionated simultaneously.
                       On execution, if the mode is isothermal or
                       ALPHAMELTS_DELTAT is zero and the mode is isobaric or
                       ALPHAMELTS_DELTAP is zero then you will be asked the
                       number of iterations you wish to perform. The program
                       will then request the file type / number of files and the
                       name(s) before the first equilibration9. The assimilant
                       bulk composition may be fixed by a single enrichment_file
                       or binary restart_file, or by providing separate
                       enrichment_files for each mineral phase in the
                       assimilant. For separate enrichment_files, phase
                       compositions are given in wt% oxides and it is up to the
                       user to ensure the solid compositions are close to
                       stoichiometric in the appropriate pure phase or solid
                       solution model (see the forum for a list of mineral phase
                       end members). Trace element concentrations should be the
                       bulk assimilant values rather than individual mineral
                       ones. A single liquid composition can be input instead
                       of, or in addition to, mineral compositions.
                       Alternatively, multiple enrichment_files can be requested
                       so that the assimilant bulk composition can be changed
                       for each iteration. To use multiple enrichment_files in
                       this way, see the ‘Table file’ section for the filename
                       format. The program assumes that the indices are reset
                       each time option 4 is called.
                       You can enter the mass of assimilant to be added at each
                       subsequent stage or, for melts_file(s) only, take the
                       value(s) from the ‘Initial Mass:’ line(s). If the mass of
                       assimilant is specified for separate mineral melts_files
                       then the value entered is the total mass and the
                       ‘Initial Mass:’ lines will be used to determine the
                       proportions of the mineral phases. Major and, if
                       appropriate, trace elements will be mixed (to mix trace
                       elements only use ALPHAMELTS_FLUX_MELTING). If trace
                       element calculations are switched on, the
                       enrichment_file(s) must contain the same trace elements
                       as the melts_file. For separate mineral files, the trace
                       elements can be included in each mineral file, to avoid
                       read errors, but only the vales from the first
                       enrichment_file will be used. If running in isothermal
                       mode, or the like, then the P, T and fO2 buffer from the
                       assimilant file will be ignored and set to the current
                       values, regardless of the input file type. For
                       isenthalpic mode: please set ALPHAMELTS_DELTAP to zero,
                       for reasons given in ‘Thermodynamic Path’, and
                       ALPHAMELTS_DELTAT to zero so that the number of
                       iterations can be controlled. We suggest you set
                       ALPHAMELTS_SAVE_ALL so that you can gradually build up
                       the number of iterations. In this mode, if you use a
                       single text enrichment_file (or files) for the assimilant
                       then the program will try to find a thermodynamically
                       equilibrated state, using a superliquidus start. If you
                       use separate mineral phase enrichment_files the
                       enthalphy of each phase is calculated at temperature
                       given in the first enrichment_file, without
                       re-equilibration; this is similar to the GUI
                       assimilation routines. If you use a binary assimilant
                       file then it will use the previously calculated starting
                       conditions but the pressure in the file must match the
                       current pressure (or the temperatures must match for
                       isochoric mode). If this is not the case an error message
                       will be printed; either (a) redo the restart_file for the
                       correct conditions or (b) use menu option 2 to set the
                       current values to those used to generate the restart_file
                       and then call option 4 again."""
        ),
    ),
    "FLUX_MELTING": mvar(
        set=False,
        type=bool,
        desc=description(
            """This variable causes a user-defined proportion of a
                         second composition, trace elements only, to be mixed
                         in after each calculation stage (see ‘Bulk
                         composition’). If used in conjunction with
                         ALPHAMELTS_CONTINUOUS_MELTING and
                         ALPHAMELTS_DO_TRACE_H2O it can simulate flux melting by
                         a hydrous fluid (to simulate fluxing by a metasomatic
                         melt use ALPHAMELTS_ASSIMILATE instead). We suggest you
                         set ALPHAMELTS_SAVE_ALL so that you can gradually build
                         up the number of iterations, e.g. until the system
                         achieves steady state, by repeated calling of the
                         ‘execute’ menu option. Execution is essentially the
                         same as for ALPHAMELTS_ASSIMILATE but for trace
                         elements only. As the mass of the enriching
                         composition is not necessarily defined, the user is
                         asked instead for the mass proportion of the new
                         composition to add (similar to ‘Source Mixer’ in menu
                         option 12). Alternatively, for melts_files(s), the
                         ‘Initial Mass:’ lines may be used; in which case the
                         old and new compositions are mixed in the ratio
                         old_mass:new_mass. Note that, as major elements are not
                         mixed, the total system mass (i.e. old_mass) will be
                         unaffected."""
        ),
    ),
    "DRY_ITER_PATIENCE": mvar(
        default=100,
        type=int,
        validator=typecast(
            int, [lambda x: isinstance(x, int), lambda x: (x >= 0) & (x <= 100)]
        ),
        desc=description(
            """If simulating flux melting or assimilation,
                              this is the maximum number of consecutive
                              iterations that alphaMELTS will run without any
                              melting occurring before it will give up and
                              return to the menu."""
        ),
    ),
    # pHMELTS and Trace Elements
    "DO_TRACE": mvar(
        set=False,
        type=bool,
        desc=description(
            """Implements attached trace element partitioning function
                     for those elements listed in the melts_file."""
        ),
    ),
    "DO_TRACE_H2O": mvar(
        set=False,
        type=bool,
        desc=description(
            """For the case where water is to be treated as a trace
                         element this option adds an iteration on the H2O
                         content, as described in Asimow, et al. [2004]."""
        ),
    ),
    "HK_OL_TRACE_H2O": mvar(
        set=False,
        type=bool,
        desc=description(
            """By default the Mosenfelder, et al. [2005] model
                            for water solubility in olivine is used. This
                            environment variable uses the Hirth and Kohlstedt
                            [1996] model instead, which gives lower solubility
                            and consequently lower partition coefficients."""
        ),
    ),
    "HK_PXGT_TRACE_H2O": mvar(
        default="mineral-melt",
        type=bool,
        validator=lambda x: x in ["mineral-melt", "mineral-mineral"],
        desc=description(
            """If the Mosenfelder, et al. [2005] model for
                              water solubility in olivine is used then by
                              default the mineral-melt partition coefficients
                              for water with orthopyroxene, clinopyroxene and
                              garnet are still those of Hirth and Kohlstedt
                              [1996]. Setting this variable to ‘mineral-mineral’
                              means the solubility of water in opx, cpx and
                              garnet are linked to the newer water solubility
                              in olivine model, which is equivalent to
                              preserving the mineral-mineral water partition
                              coefficients of Hirth and Kohlstedt [1996].
                              Setting it to ‘mineral-melt’ retains the default
                              behavior."""
        ),
    ),
    "2X_OPX_TRACE_H2O": mvar(
        set=False,
        type=bool,
        desc=description(
            """By default the solubility of water in
                             clinopyroxene is twice that in orthopyroxene,
                             based on the results of Hirth and Kohlstedt [1996].
                             If this variable is set then the solubility of
                             water in opx is scaled up by to be equal to the
                             cpx value, consistent with the observations of
                             Hauri et al. [2006]. This option can be used
                             regardless of whether the Mosenfelder, et al.
                             [2005] or Hirth and Kohlstedt [1996] models are
                             being used for olivine or opx, cpx and garnet. For
                             example, if either ALPHAMELTS_HK_OL_TRACE_H2O or
                             ALPHAMELTS_HK_PXGT_TRACE_H2O is set then the
                             mineral-mineral partition coefficients will be
                             those from Table 1 of Hirth and Kohlstedt [1996],
                             except that olivine-opx value of 0.2 will be
                             replaced by 0.1."""
        ),
    ),
    "TRACE_DEFAULT_DPTX": mvar(
        set=False,
        type=bool,
        desc=description(
            """By default all partition coefficients used in
                               trace element calculations are constant. If
                               ALPHAMELTS_TRACE_VARIABLE_D is set then D =
                               D(P,T,X) are calculated for elements and phases
                               in Table 1. Constant partition coefficients will
                               be used for all other elements / phases. This
                               list may be modified in the trace_data_file, as
                               previously explained."""
        ),
    ),
    "TS_TRACE_NORMALIZATION": mvar(
        set=False,
        type=int,
        validator=typecast(
            int, [lambda x: isinstance(x, int), lambda x: (x >= 1) & (x <= 4)]
        ),
        desc=description(
            """If set, this chooses one of four
                                   compositions to normalize trace elements to
                                   (if any):
                                   1. PM Sun and McDonough [1989];
                                   2. DMM Workman and Hart [2005];
                                   3. PM McKenzie and O'Nions [1991; 1995];
                                   4. DM of McKenzie and O'Nions [1991; 1995].
                                   Sample input files showing concentrations for
                                   each of the idealized source compositions
                                   above are provided as illustrations; note
                                   that some elements (i.e. Ni, Cr, and Mn)
                                   appear as major and trace elements because
                                   their inclusion in the liquid calibration
                                   differs between MELTS and pMELTS. Isotopes
                                   are normalized to the ‘non-isotope’ abundance
                                   (see the ‘MELTS file’ section). This option
                                   is useful if the source composition given in
                                   the melts_file is different from the four
                                   options above. If you wish to normalize to
                                   the source in the melts_file the simplest
                                   thing is to provide the list of elements with
                                   ‘1.0’ for each of the abundances (except
                                   isotopes)."""
        ),
    ),
    "TRACE_INPUT_FILE": mvar(
        set=False,
        desc=description(
            """Gives the name of the trace_data_file, which may
                             be used to change partition coefficients, determine
                             whether variable partition coefficients are
                             calculated and with which parameters, as described
                             above."""
        ),
    ),
    "TRACE_USELIQFEMG": mvar(
        set=False,
        type=bool,
        desc=description(
            """By default the Mg# of the melt, needed to
                             estimate D(P,T,X) for clinopyroxene, is estimated
                             from the clinopyroxene composition using Equation
                             35 from Wood and Blundy [1997]). If this
                             environment variable is set then the Mg# of the
                             melt is taken directly from the (pH)MELTS
                             calculated liquid composition. If no liquid is
                             present then the program will revert to the default
                             behavior and use the clinopyroxene composition."""
        ),
    ),
    # Input options
    "ALPHAMELTS_ADIABAT_BIN_FILE": mvar(
        set=False,
        desc=description(
            """Unavoidable changes within the
                                        programs mean that binary restart_files
                                        created with version 2.0+ of Adiabat_1ph
                                        will not read into alphaMELTS. Once you
                                        have read in the file, if you
                                        immediately save it again it will be
                                        updated to Adiabat_1ph 3.0 format.
                                        Adiabat_1ph 3.0 format is compatible
                                        with alphaMELTS 1.0+ so unset this
                                        environment variable for subsequent
                                        runs. For more details see
                                        ADIABAT_V20_BIN_FILE in the
                                        Adiabat_1ph 3 documentation."""
        ),
    ),
    # Output options
    "CELSIUS_OUTPUT": mvar(
        set=False,
        type=bool,
        desc=description(
            """By default, temperature input to alphaMELTS is in
                           °C, whereas temperature output is in Kelvin. When
                           this environment variable is set, text file and
                           screen output is also in °C."""
        ),
    ),
    "SAVE_ALL": mvar(
        set=False,
        type=bool,
        desc=description(
            """By default, the main output file is only written for
                     calculations made in the most recent call to menu option 4.
                     If this variable is set then results from all calculations
                     are saved and output. This provides a simple way to record
                     single calculations (menu option 3) or to build up results
                     from multiple iterations or multiple melts_files. Note that
                     even if the appropriate environment variables are set, no
                     solid or liquid fractionation will occur until menu option
                     4 is run. If using ‘Amoeba’ (menu option 17) the
                     ALPHAMELTS_SAVE_ALL function will be temporarily switched
                     off (so that iterations of the search algorithm are not
                     saved); when it is finished a single calculation (menu
                     option 3) may be used to write output for the best-fit
                     parental melt."""
        ),
    ),
    "SKIP_FAILURE": mvar(
        set=False,
        type=bool,
        desc=description(
            """Normally failure of the minimization routines means
                         that the alphamelts executable must be restarted to
                         clear the memory. If this environment variable is set
                         then a copy of the thermodynamic state is made each
                         time a successful calculation is made. If the next
                         calculation fails, this last good state (or the bulk
                         composition and starting conditions from the
                         melts_file, if it is the first calculation) is used
                         for subsequent attempts. This may be useful if you are
                         very close to a reaction when the algorithms may have
                         trouble deciding whether to add a phase to the
                         assemblage and need to overstep just slightly. It is
                         also helpful when trying out starting solutions using
                         options 2 and 3. If ALPHAMELTS_SAVE_ALL is also set,
                         the last good state will also be written to the output
                         files as a placeholder; it should be obvious that this
                         represents a skipped failure as all values, including
                         pressure and temperature, will be identical to the
                         previous output.."""
        ),
    ),
    "FAILED_ITER_PATIENCE": mvar(
        default=10,
        type=int,
        validator=typecast(
            int, [lambda x: isinstance(x, int), lambda x: (x >= 0) & (x <= 10)]
        ),
        desc=description(
            """If the minimization routines do not recover
                                 in the next few iterations then one drawback of
                                 ALPHAMELTS_SKIP_FAILURE is that it can lead to
                                 infinite loops of failure or, if triggered
                                 repeatedly, eventually to a segmentation fault.
                                 If ALPHAMELTS_SKIP_FAILURE is set, this
                                 variable is the maximum number of consecutive
                                 failed iterations that alphaMELTS will run
                                 before a clean return to the menu where
                                 parameters, such as the choice of fO2 buffer,
                                 can be adjusted or alphaMELTS completely
                                 restarted. If running in PTgrid or TPgrid
                                 modes, it will be the maximum number of failed
                                 iterations before alpahMELTS moves to the next
                                 internal loop (T-loop or P-loop
                                 respectively)."""
        ),
    ),
    "INTEGRATE_FILE": mvar(
        set=False,
        type=bool,
        desc=description(
            """Normally the integrated_output_file can be written
                           immediately after execution finishes, but if
                           alphamelts tends to crash before the menu option 16
                           can be called then memory contents will be lost. As a
                           precaution, ALPHAMELTS_INTEGRATE_FILE should be set
                           to record details of the packets of melt extracted
                           during melting to a file. If run_alphamelts.command
                           is restarted with the same settings and menu option
                           16 is called before execution then the resulting file
                           will be used as input for the melt integration. Note
                           that ALPHAMELTS_INTEGRATE_FILE is not the
                           integrated_output_file itself but an intermediate
                           file that may be used in its generation."""
        ),
    ),
    "LATENT_HEAT": mvar(
        set=False,
        type=bool,
        desc=description(
            """Normally text file output is done just after
                        equilibration and before any fractionation. If
                        ALPHAMELTS_LATENT_HEAT to set to ‘true’ alphamelts will
                        also output the state just after fractionation and just
                        before the next equilibration (when the temperature has
                        been lowered and the system cooled but not allowed to
                        crystallise any more yet). This gives a simple way to
                        calculate the latent heat of crystallization, without
                        having to do the cooling adjustment described in Ghiorso
                        [1997]. Note that, for the same MORB composition, the
                        results will differ from Figure 5 of Ghiorso [1997].
                        This apparent discrepancy is due to changes in the MELTS
                        software since it was first released and not to
                        differences in how the variation in latent heat is
                        accessed."""
        ),
    ),
    "QUICK_OUTPUT": mvar(
        set=False,
        type=bool,
        desc=description(
            """If this environment variable is set then
                         equilibrated thermodynamic states from previous
                         calculations are not saved, except as required for
                         ALPHAMELTS_SKIP_FAILURE etc. For calculations
                         comprising a large number of iterations, such as those
                         on a closely spaced P-T grid, setting this variable can
                         significantly reduce memory usage and therefore
                         increase computational speed as the calculation
                         proceeds. Reporting of fO2 will be affected slightly.
                         The Phase_mass_tbl.txt and Phase_vol_tbl.txt files will
                         not be written, which will cause
                         run_alphamelts.command to issue a warning about not
                         being able to find these files; the warning can be
                         safely ignored."""
        ),
    ),
    # Others
    "MULTIPLE_LIQUIDS": mvar(
        set=False,
        type=bool,
        desc=description(
            """This turns on exsolution of immiscible liquids. The
                             solvi are not very well determined, and we do not
                             recommend serious use of this feature, but in some
                             cases operation inside an unrecognized two-liquid field
                             can lead to path-dependent non-unique equilibria. This
                             option should not be used if trace element calculations
                             are enabled."""
        ),
    ),
    "FRACTIONATE_SECOND_LIQUID": mvar(
        set=False,
        type=bool,
        desc=description(
            """When running with
                                      ALPHAMELTS_MULTIPLE_LIQUIDS set, treat all
                                      liquids except the first as fractionating
                                      phases and remove them from the system
                                      after each equilibration."""
        ),
    ),
    "FOCUS": mvar(
        set=False,
        type=bool,
        desc=description(
            """Option to do the focusing calculation described in Asimow
                  and Stolper [1999]. It works by multiplying the mass of liquid
                  in system by a fixed factor after each equilibration."""
        ),
    ),
    "FOCUS_FACTOR": mvar(
        set=False,
        type=float,
        validator=typecast(float, numeric_validator),
        desc=description(
            """When ALPHAMELTS_FOCUS is set, this determines the
                         multiplication factor for the mass of liquid. Usually
                         it will be a number slightly greater than unity, like
                         the 100th root of 2. If ALPHAMELTS_FOCUS is set without
                         a value in ALPHAMELTS_FOCUS_FACTOR the focusing
                         calculation loop will be ignored."""
        ),
    ),
    "INTEGRATE_PHI": mvar(
        set=False,
        type=bool,
        desc=description(
            """Normally, packets of extracted melt are saved in
                          memory (and, optionally, written out to the
                          ALPHAMELTS_INTEGRATE_FILE ‘crash’ file, described
                          above) by mass i.e. in terms of dF. If this variable
                          is set then the packets of extracted melt will be
                          saved by volume i.e. in terms of dPhi. As the full
                          pressure integral (calculation of crustal thickness
                          etc.) is not strictly correct for increments in Phi,
                          that option will not be offered. However, the
                          opportunity to interpolate for equal increments of
                          melt fraction by volume will be available."""
        ),
    ),
}



import pandas as pd
import numpy as np
import pyrolite.geochem
from ..util.log import Handle

logger = Handle(__name__)


def phasename(phaseID):
    """
    Take a phase ID and return the name of the phase.

    Parameters
    ------------
    phaseID : :class:`str`
        ID for the particular phase (e.g. 'olivine_0')

    Returns
    --------
    :class:`str`
        Name of the phase.
    """
    if phaseID.find("_") > 0:
        n = phaseID[: phaseID.find("_")]
    else:
        n = phaseID
    return n


def tuple_reindex(df, columns=["pressure", "temperature"]):
    """
    Create an index based on tuples from multiple columns.

    Parameters
    -----------
    df: :class:`pandas.DataFrame`
        Table DataFrame to reindex.
    columns : :class:`list`
        List of columns to incorporate into the tuple index.

    Returns
    -------
    :class:`pandas.DataFrame`
        Reindexed DataFrame.
    """
    df.index = df.loc[:, columns].astype(int).itertuples(index=False)
    return df


def integrate_solid_composition(df, frac=True):
    """
    Integrate solid compositions to return a 'cumulate' like
    composition. Note that in the case of non-fractional crystallisation
    this will correspond to the solid composition.

    Parameters
    -----------
    df : :class:`pandas.DataFrame`
        DataFrame to integrate.
    frac : :class:`bool`
        Whether the experiment is a fractional crystallisation experiment.

    Returns
    -----------
    df : :class:`pandas.DataFrame`
        DataFrame containing an integrated solid composition.
    """
    assert not "experiment" in df.columns, "Designed for single tables."
    slds = df.loc[df.phase == "solid", :]
    idx = (
        df.loc[:, ["pressure", "temperature", "step"]]
        .dropna()
        .drop_duplicates()
        .sort_values("step")
    )
    if frac:
        cumulate = pd.DataFrame(columns=slds.columns, index=idx.index)
        # solids typically don't exist for part of the history, so we need reindex here
        # rather than .loc[<index list>, :]
        cumulate["mass"] = np.nancumsum(slds["mass"].reindex(index=idx.index).values)

        chem = slds.reindex(
            index=idx.index,
            columns=[
                i for i in slds.pyrochem.list_compositional if i not in ["S", "H", "V"]
            ],
        )
        chem = chem.apply(pd.to_numeric, errors="coerce")
        increments = (
            slds["mass"].reindex(index=idx.index).values[:, np.newaxis] * chem.values
        )
        cumulate[chem.columns] = np.nancumsum(increments, axis=1)
        cumulate[["pressure", "temperature", "step"]] = slds.loc[
            :, ["pressure", "temperature", "step"]
        ]
    else:
        cumulate = slds.reindex(index=idx.index)
    cumulate.pyrochem.add_MgNo()
    return cumulate


def integrate_solid_proportions(df, frac=True):
    """
    Integrate solid proportions to return a 'cumulate' split by integrated phase
    masses. Note that in the case of non-fractional crystallisation
    this will correspond to the overall solid phase abundances.

    Parameters
    -----------
    df : :class:`pandas.DataFrame`
        DataFrame to integrate.
    frac : :class:`bool`
        Whether the experiment is a fractional crystallisation experiment.

    Returns
    -----------
    df : :class:`pandas.DataFrame`
        DataFrame containing integrated solid phase proportions.
    """
    assert not "experiment" in df.columns, "Designed for single tables."
    # another dataframe for integrated minerals
    phaseIDs = sorted(
        [
            pID
            for pID in df.phaseID.unique()
            if (not pd.isnull(pID)) and ("liquid" not in pID)
        ]
    )
    idx = (
        df.loc[:, ["pressure", "temperature", "step"]]
        .dropna()
        .drop_duplicates()
        .sort_values("step")
    )
    # empty dataframe
    mindf = pd.DataFrame(
        columns=["pressure", "temperature", "step"] + phaseIDs, index=idx.index
    )
    for p in phaseIDs:  # integrate cumulate mass per phase
        # mindf should have all of the mineral index values
        mindf.loc[df.loc[df.phaseID == p, "mass"].index.values, p] = df.loc[
            df.phaseID == p, "mass"
        ].values
    mindf = mindf.loc[idx.index, :]  # sort index
    if frac:
        mindf = mindf.apply(np.nancumsum, axis=0)  # accumulate minerals
    # fractioal mass of total cumulate
    mindf = mindf.div(mindf.sum(axis=1).replace(0, np.nan), axis=0) * 100.0
    PTS = idx
    mindf.loc[idx.index, ["pressure", "temperature", "step"]] = PTS

    mindf = mindf.fillna(0)
    return mindf



"""
Functions for loading alphaMELTS tables to :class:`pandas.DataFrame`.

Notes
-------

    * Some variable abbreviations are converted to their names (e.g.
        enthalpy, entropy and volume).
"""
import re
import io
import json
import pandas as pd
import numpy as np
from pathlib import Path
from pyrolite.util.pd import zero_to_nan
from ..parse import from_melts_cstr
from ..util.tables import (
    phasename,
    tuple_reindex,
    integrate_solid_composition,
    integrate_solid_proportions,
)
from ..util.log import Handle

logger = Handle(__name__)

TABLES = {
    "Phase_mass_tbl.txt",
    "Phase_vol_tbl.txt",
    "System_main_tbl.txt",
    "Liquid_comp_tbl.txt",
    "Solid_comp_tbl.txt",
    "Bulk_comp_tbl.txt",
    "Trace_main_tbl.txt",
    "Phase_main_tbl.txt",
}

THERMO = {"H": "enthalpy", "S": "entropy", "V": "volume"}
THERMO.update({c: c.lower() for c in ["Pressure", "Temperature"]})


def convert_thermo_names(df):
    """
    Convert the abbreviations for thermodynamic variables (e.g. H, S) to
    expanded names. This avoids some issues with pulling out compositions.

    Parameters
    ------------
    :class:`pandas.DataFrame`
        DataFrame to convert column names for.

    Returns
    -------
    :class:`pandas.DataFrame`
        DataFrame with table information.
    """
    df.rename(columns=THERMO, inplace=True)
    return df


def read_phase_table(tab):
    """
    Import a phase table to a dataframe.

    Parameters
    ------------
    tab : :class:`str`
        String containing the table to be imported, with its title.

    Returns
    -------
    :class:`pandas.DataFrame`
        DataFrame with phase table information.
    """
    lines = [i for i in re.split(r"[\n\r]", tab) if i]
    phaseID = lines[0].split()[0].strip()
    headers = lines[1].strip().split()
    linelen = [len(l.strip().split()) for l in lines[1:]]
    if not all([l == linelen[0] for l in linelen]):
        if linelen[0] == (linelen[1] - 1):  # known errors for neph, kals, alloy
            if any(
                [
                    phase in phaseID
                    for phase in ["nepheline", "kalsilite", "alloy-solid"]
                ]
            ):  # inconsistent headers
                expect = [
                    "Pressure",
                    "Temperature",
                    "mass",
                    "S",
                    "H",
                    "V",
                    "Cp",
                    "structure",
                    "formula",
                ]
                headers = [i for i in expect] + [i for i in headers if i not in expect]
        else:
            logger.warning(  # unkonwn line length error
                "Inconsistent line lengths for {} table: {}".format(
                    phaseID, " ".join([str(i) for i in linelen])
                )
            )
    buff = io.BytesIO("\n".join([" ".join(headers)] + lines[2:]).encode("UTF-8"))
    table = pd.read_csv(buff, sep=" ")
    table["phaseID"] = phaseID
    table["phase"] = phasename(phaseID)
    return table


def read_melts_tablefile(filepath, kelvin=False, skiprows=3, **kwargs):
    """
    Read a melts table (a space-separated value file).

    Parameters
    -----------
    filepath : :class:`str` | :class:`pathlib.Path`
        Filepath to the melts table.
    kelvin : :class:`bool`
        Whether the imported table has temperature listed in kelvin.
    skiprows : :class:`int`
        Number of rows above the table headers.

    Returns
    -------
    :class:`pandas.DataFrame`
        DataFrame with table information.
    """
    path = Path(filepath)
    with open(path) as tab:
        lines = [i for i in tab.readlines()[skiprows:] if i]
        headers = lines[0].strip().split()
        for ix, h in enumerate(headers):
            if (
                headers[: ix + 1].count(h) > 1
            ):  # logfO2(absolute) is sometimes duplicated
                headers[ix] = h + ".1"  # silence duplicate
        linelen = [len(l.strip().split()) for l in lines]
        if not all([l == linelen[0] for l in linelen]):
            logger.debug(  # debug here because these tables are often left-empty
                "Inconsistent line lengths for table: {}".format(
                    "-".join([str(i) for i in linelen])
                )
            )
    buff = io.BytesIO("".join(lines[1:]).encode("UTF-8"))
    df = pd.read_csv(buff, sep=" ", names=headers, **kwargs)
    df = df.loc[
        :, ~df.columns.str.replace("(\.\d+)$", "").duplicated()
    ]  # remove duplicate columns
    df = df.dropna(how="all", axis=1)

    df = convert_thermo_names(df)

    if ("temperature" in df.columns) and not kelvin:
        df["temperature"] -= 273.15
    if ("MgO" in df.columns) and ("FeO" in df.columns):
        # should update this to be if there's both iron and magnesium species
        df.pyrochem.add_MgNo()
    df = zero_to_nan(df)
    df = tuple_reindex(df)
    return df


def phasetable_from_alphameltstxt(filepath, kelvin=False):
    """
    Read the phasemain file into a single table. Note that the alphaMELTS table
    includes all other tables also (except for traces).

    Parameters
    ------------
    filepath : :class:`str` | :class:`pathlib.Path`
        Filepath to the melts table.
    kelvin : :class:`bool`
        Whether the exported table has temperature listed in kelvin.

    Returns
    -------
    :class:`pandas.DataFrame`
        DataFrame with table information.

    Todo
    ------

    * If this file exists and is properly formatted, we can pull in alomst all
        the data here.
    """

    filepath = Path(filepath)
    assert filepath.exists()
    with open(str(filepath), "r") as f:
        df = pd.DataFrame()
        data = f.read().strip()
        tables = re.split("Title: ", data)
        tables = [t for t in re.split(r"Title: .*[\n\r][\n\r]+", data, re.DOTALL) if t]
        phasetbl = [t for t in tables if t[0] == t[0].lower()]
        if len(phasetbl) != 1:
            logger.warning("Imported alphaMELTS_tbl.txt incorrectly formatted.")
        else:
            phasetbl = phasetbl[0]
            phasettlbs = re.split(r"[\n\r][\n\r]+", phasetbl.strip())
            for tab in phasettlbs:
                tabdf = read_phase_table(tab)
                df = df.append(tabdf, sort=False)

    df = convert_thermo_names(df)
    non_num = ["step", "structure", "phaseID", "phase", "formula"]
    num = [i for i in df.columns if i not in non_num]
    df[num] = df[num].apply(pd.to_numeric, errors="coerce")

    if "formula" in df.columns:
        df.loc[:, "formula"] = df.loc[:, "formula"].apply(from_melts_cstr)

    if ("temperature" in df.columns) and not kelvin:
        df["temperature"] -= 273.15

    if ("MgO" in df.columns) and ("FeO" in df.columns):
        # should update this to be if there's both iron and magnesium species
        df.pyrochem.add_MgNo()

    df = zero_to_nan(df)
    df = tuple_reindex(df)
    return df


def phasetable_from_phasemain(filepath, kelvin=False):
    """
    Read the phasemain file into a single table.

    Parameters
    ------------
    filepath : :class:`str` | :class:`pathlib.Path`
        Filepath to the melts table.
    kelvin : :class:`bool`
        Whether the exported table has temperature listed in kelvin.

    Returns
    -------
    :class:`pandas.DataFrame`
        DataFrame with table information.
    """
    df = pd.DataFrame()
    with open(str(filepath)) as f:
        data = re.split(r"[\n\r][\n\r]+", f.read())[1:]  # double line sep
        for tab in data:
            df = df.append(read_phase_table(tab), sort=False)

    df = convert_thermo_names(df)
    non_num = ["step", "structure", "phaseID", "phase", "formula"]
    num = [i for i in df.columns if i not in non_num]
    df[num] = df[num].apply(pd.to_numeric, errors="coerce")

    if "formula" in df.columns:
        df.loc[:, "formula"] = df.loc[:, "formula"].apply(from_melts_cstr)

    if ("temperature" in df.columns) and not kelvin:
        df["temperature"] -= 273.15

    if ("MgO" in df.columns) and ("FeO" in df.columns):
        # should update this to be if there's both iron and magnesium species
        df.pyrochem.add_MgNo()

    df = zero_to_nan(df)
    df = tuple_reindex(df)
    return df


def import_tables(pth, kelvin=False):
    """
    Import tables from a directory.

    Parameters
    -----------
    kelvin : :class:`bool`
        Whether to keep temperatures in kelvin.

    Returns
    --------
    system : :class:`pandas.DataFrame`

    phases : :class:`pandas.DataFrame`
    """
    sysfile, bulkfile, solidfile, alphafile = [
        pth / t
        for t in [
            "System_main_tbl.txt",
            "Bulk_comp_tbl.txt",
            "Solid_comp_tbl.txt",
            "alphaMELTS_tbl.txt",
        ]
    ]
    try:
        for f in [sysfile, bulkfile, solidfile, alphafile]:
            assert f.exists()
    except AssertionError as err:
        msg = "File missing from {}: {}".format(
            pth, ", ".join([i.name for i in pth.iterdir()])
        )
        raise FileNotFoundError(msg) from err
    # system table
    system = read_melts_tablefile(sysfile, skiprows=3, kelvin=kelvin)
    system["step"] = np.arange(system.index.size)  # generate the step index
    system["mass%"] = (system["mass"] / system["mass"].values[0]) * 100
    system["volume%"] = (system["volume"] / system["volume"].values[0]) * 100
    system = system.reindex(
        columns=["step"] + [i for i in system.columns if i != "step"]
    )

    phase = phasetable_from_alphameltstxt(pth / "alphaMELTS_tbl.txt", kelvin=kelvin)
    bulk = read_melts_tablefile(pth / "Bulk_comp_tbl.txt", skiprows=3, kelvin=kelvin)
    solid = read_melts_tablefile(pth / "Solid_comp_tbl.txt", skiprows=3, kelvin=kelvin)

    phase["step"] = system.loc[phase.index, "step"]
    bulk["step"] = system.loc[bulk.index, "step"]
    solid["step"] = system.loc[solid.index, "step"]

    bulk["phase"] = "bulk"
    solid["phase"] = "solid"

    solid = solid.loc[solid["mass"] > 0.0, :]  # drop where no solids present
    # traces could be imported here

    for tb in [bulk, solid]:
        phase = phase.append(tb, sort=False)
    # integrated solids for fractionation - if the system mass changes significantly
    # could add this threshold as a parameter
    frac = system.mass.max() / system.mass.min() > 1.05
    cumulate_comp = integrate_solid_composition(phase, frac=frac)
    cumulate_comp["phase"] = "cumulate"
    phase = phase.append(cumulate_comp, sort=False)

    cumulate_phases = integrate_solid_proportions(phase, frac=frac)
    cumulate_comp["phase"] = "cumulate"
    # TODO: Integrate these?

    phase["step"] = system.loc[phase.index, "step"]
    phase = phase.reindex(columns=["step"] + [i for i in phase.columns if i != "step"])

    phase["mass%"] = phase["mass"] / system.loc[phase.index, "mass"].values[0] * 100
    phase["volume%"] = (
        phase["volume"] / system.loc[phase.index, "volume"].values[0] * 100
    )
    # Convert column dtypes where necessary
    obj_columns = ["phaseID", "phase", "formula", "structure"]
    numeric_columns = [col for col in phase.columns if col not in obj_columns]
    phase[numeric_columns] = phase[numeric_columns].apply(
        pd.to_numeric, errors="coerce"
    )
    return system, phase


def import_batch_config(filepath):
    """
    Import a batch configuration file.

    Parameters
    -----------
    filepath : :class:`str` | :class:`pathlib.Path`

    Returns
    ---------
    :class:`dict`
        Configuration dictionary indexed by hashes.
    """

    filepath = Path(filepath)
    if filepath.is_dir():
        # find config
        cfgpath = filepath / "meltsBatchConfig.json"
    else:
        cfgpath = filepath.with_suffix("json")

    with open(str(cfgpath), "r") as f:
        cfg = json.loads(f.read())
    return cfg


def aggregate_tables(
    lst=Path("./"), kelvin=False, validate_path=lambda x: len(x.name) == 10
):
    """
    Aggregate a number of melts tables to a single dataframe.

    Parameters
    ------------
    lst : :class:`str` | :class:`pathlib.Path` | :class:`list`
        Directory, list of directories or list of 2-dataframe tuples.
    kelvin : :class:`bool`
        Whether to keep temperatures in kelvin.
    validate_path :
        Function to validate path names.

    Parameters
    ------------
    system : :class:`pandas.DataFrame`
        System aggregate table.
    phases : :class:`pandas.DataFrame`
        Phases aggregate table.
    """
    if isinstance(lst, (str, Path)):
        # if the input is a directory, aggregate subfolders
        lst = [x for x in Path(lst).rglob("*") if (x.is_dir() and validate_path(x))]

    system, phases = pd.DataFrame(), pd.DataFrame()
    if isinstance(lst[0], (str, Path)):
        # if the list is of filenames, aggregate the tables one by one
        for d in lst:
            try:
                S, P = import_tables(d, kelvin=kelvin)
                # ensure the experiment name is incorporated
                S["experiment"] = d.name
                P["experiment"] = d.name

                system = system.append(S, sort=False)
                phases = phases.append(P, sort=False)
            except Exception as e:
                logger.warning("{} at {}.".format(e, d.name))  # record the error
    elif isinstance(lst[0], (list, tuple)) and isinstance(lst[0][0], (pd.DataFrame)):
        # if the list is of tuples of dataframes,
        # aggregate them to a single table
        Sagg, Pagg = pd.DataFrame(), pd.DataFrame()
        for ix, d in enumerate(lst):

            S, P = d
            # ensure the experiment index is incorporated
            S["experiment"] = ix
            P["experiment"] = ix

            system = system.append(S, sort=False)
            phases = phases.append(P, sort=False)
    else:
        raise NotImplementedError

    system = system.reindex(
        columns=["experiment"] + [i for i in system.columns if i != "experiment"]
    )
    phases = phases.reindex(
        columns=["experiment"] + [i for i in phases.columns if i != "experiment"]
    )
    return system, phases



from pathlib import Path
from ..parse import read_envfile, read_meltsfile
from ..util.log import Handle

logger = Handle(__name__)


def make_meltsfolder(
    name, title=None, meltsfile=None, indir=None, env="./alphamelts_default_env.txt"
):
    """
    Create a folder for a given meltsfile, including the default environment file.
    From this folder, pass these to alphamelts with
    :code:`run_alphamelts.command -m <meltsfile> -f <envfile>`.

    Parameters
    -----------
    name : :class:`str`
        Name of the folder.
    title : :class:`str`
        Title of the experiment. This will be the meltsfile name.
    meltsfile : :class:`str`
        String containing meltsfile info.
    indir : :class:`str` | :class:`pathlib.Path`
        Path to the base directory to create melts folders in.
    env : :class:`str` | :class:`pathlib.Path`
        Path to a specific environment file to use as the default environment for the
        experiment.

    Returns
    --------
    :class:`pathlib.Path`
        Path to melts folder.

    Todo
    ------
        * Options for naming environment files
    """
    if indir is None:
        indir = Path("./")
    else:
        indir = Path(indir)
    assert meltsfile is not None
    name = str(name)  # need to pathify this!
    title = title or name
    title = str(title)  # need to pathify this!
    experiment_folder = indir / name
    if not experiment_folder.exists():
        experiment_folder.mkdir(parents=True)

    meltsfile, mpath = read_meltsfile(meltsfile)
    assert experiment_folder.exists() and experiment_folder.is_dir()
    (experiment_folder / title).with_suffix(".melts").touch()
    with open(str((experiment_folder / title).with_suffix(".melts")), "w") as f:
        f.write(meltsfile)

    (experiment_folder / "environment").with_suffix(".txt").touch()
    env, epath = read_envfile(env, unset_variables=False)
    with open(str(experiment_folder / "environment.txt"), "w") as f:
        f.write(env)

    return experiment_folder  # return the folder name



"""
alphaMELTS download and installsation utilities.
"""
import os, sys, platform
import subprocess, shutil
import io
import requests
import zipfile
from pathlib import Path
from pyrolite.util.general import copy_file, extract_zip, remove_tempdir
from pyrolite.util.web import internet_connection
from .util.general import pyrolite_meltsutil_datafolder, check_perl
import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())
logger = logging.getLogger(__name__)


def download_melts(directory, version=None):
    """
    Download and extract melts zip file to a given directory.

    Parameters
    ----------
    directory :  :class:`str` |  :class:`pathlib.Path`
        Directory into which to extract melts.
    version : :class:`str`
        Which alphamelts version to use. Defaults to latest stable version.

    Todo
    -----
        * Check version, enable update-style overwrite
    """
    try:
        assert internet_connection()
        system = platform.system()
        release = platform.release()
        platver = platform.version()
        bits, linkage = platform.architecture()
        bits = bits[:2]
        if version is None:
            version = 1.9  # eventually should update this programatically!
        mver = "-".join(str(version).split("."))
        zipsource = "https://magmasource.caltech.edu/alphamelts/zipfiles/"
        if system == "Linux":
            if ("Microsoft" in release) or ("Microsoft" in platver):
                url = zipsource + "wsl_alphamelts_{}.zip".format(mver)
            else:
                url = zipsource + "linux_alphamelts_{}.zip".format(mver)
        elif system == "Darwin":
            url = zipsource + "macosx_alphamelts_{}.zip".format(mver)
        elif system == "Windows":
            url = zipsource + "windows_alphamelts_{}.zip".format(mver)
            install_file = "alphamelts_win{}.exe".format(bits)
        else:
            raise NotImplementedError("System unknown: {}".format(system))

        # Set install directory for .bat files
        directory = Path(directory)
        if directory:
            install_dir = directory
        else:
            install_dir = "."

        if not install_dir.exists():
            install_dir.mkdir(parents=True)

        r = requests.get(url, stream=True)
        if r.ok:
            z = zipfile.ZipFile(io.BytesIO(r.content))
            extract_zip(z, install_dir)
    except AssertionError:
        raise AssertionError("Need an internet connection to download.")


def install_melts(
    install_dir=None,
    link_dir=None,
    eg_dir=None,
    native=True,
    temp_dir=Path("~").expanduser() / "temp" / "temp_melts",
    keep_tempdir=False,
    with_readline=True,
    local=False,
    version=None,
):
    """
    Parameters
    ----------
    install_dir : :class:`str` | :class:`pathlib.Path`
        Directory into which to install melts executable.
    link_dir : :class:`str` | :class:`pathlib.Path`
        Directory into which to deposit melts links.
    eg_dir : :class:`str` | :class:`pathlib.Path`
        Directory into which to deposit melts examples.
    native : :class:`bool`, :code:`True`
        Whether to install using python (:code:`True`) or the perl scripts (windows).
    temp_dir : :class:`str` | :class:`pathlib.Path`, :code:`$USER$/temp/temp_melts`
        Temporary directory for melts file download and install.
    keep_tempdir : :class:`bool`, :code:`False`
        Whether to cache tempoary files and preserve the temporary directory.
    with_readline : :class:`bool`, :code:`True`
        Whether to also attempt to install with_readline.
    local : :class:`bool`
        Whether to install a version of melts into an auxiliary pyrolite data folder.
        This will override
    """
    system = platform.system()
    platrel = platform.release()
    platver = platform.version()
    bits, linkage = platform.architecture()
    bits = bits[:2]

    temp_dir = Path(temp_dir)

    if (
        temp_dir / "install.command"
    ).exists() and version is None:  # already downloaded for some reason
        pass
    else:
        logger.info("Downloading Melts")
        if not temp_dir.exists():
            temp_dir.mkdir(parents=True)
        download_melts(temp_dir, version=version)

    if local:
        install_dir = pyrolite_meltsutil_datafolder(subfolder="localinstall")
    else:
        install_dir = Path(install_dir)

    if not install_dir.exists():
        install_dir.mkdir(parents=True)

    if (link_dir is not None) and (not local):
        link_dir = Path(link_dir)
    else:
        link_dir = install_dir / "links"

    if not link_dir.exists():
        link_dir.mkdir(parents=True)

    if (eg_dir is not None) and (not local):
        eg_dir = Path(eg_dir)
    else:
        eg_dir = install_dir / "examples"

    if not eg_dir.exists():
        eg_dir.mkdir(parents=True)

    logger.info("Installing to {} from {}".format(install_dir, temp_dir))
    try:
        if check_perl() and (not native):
            """
            Notes
            -----
            Setting an install folder other than the download folder
            seems to fail here.
            Melts gets confused with the directory structure...
            and creates .bat files which point to the wrong place
            """
            install_source = os.path.join(str(temp_dir), "install.command")
            args = ["perl", install_source]

            # [C:\Users\<>\Documents\bin]
            # [./\examples]
            # use default settings file
            # continue
            # return to finish
            inputs = ["", str(link_dir), str(eg_dir), "", "y", "", ""]
            p = subprocess.run(
                args, input=("\n".join(inputs)).encode("UTF-8"), stdout=subprocess.PIPE
            )

            logger.info("\n" + p.stdout.decode("UTF-8"))
            assert p.returncode == 0

            # copy files from tempdir to install_dir
            regs = []  #'command', 'command_auto_file', 'path', 'perl']
            comms = ["install", "column_pick", "file_format", "run_alphamelts"]
            for (prefixes, ext) in [(regs, ".reg"), (comms, ".command")]:
                for prefix in prefixes:
                    temp_regpath = (temp_dir / prefix).with_suffix(ext)
                    install_regpath = install_dir / temp_regpath.name

                    shutil.copy(str(temp_regpath), str(install_regpath))
        elif native:
            # need to split into platforms
            egs = []
            for g in ["*.melts", "*.txt", "*.m ", "*.pdf"]:
                egs += list(temp_dir.glob(g))
            comms = [
                "column_pick.command",
                "file_format.command",
                "run_alphamelts.command",
            ]

            # getting the executable file
            if system == "Windows":
                alphafile = "alphamelts_win{}.exe".format(bits)
                alias = alphafile.replace(".exe", "")
            elif system == "Linux":
                if ("Microsoft" in platrel) or ("Microsoft" in platver):
                    alphafile = "alphamelts_wsl"
                    # with_readline
                else:
                    alphafile = "alphamelts_linux{}".format(bits)
                    # with_readline
            elif system == "Darwin":
                alphafile = "alphamelts_macosx{}".format(bits)
                # with_readline

            # copy executable; alias is used on windows to make sure run_command links to the exe
            alphaalias = [alphafile, "alphamelts"][system == "Windows"]
            copy_file(temp_dir / alphafile, install_dir / alphaalias, permissions=0o777)
            # copy examples
            for (target, files) in [(eg_dir, egs)]:
                for fn in files:
                    copy_file(temp_dir / fn.name, target / fn.name)

            # copycommand files
            for (target, files) in [(install_dir, [(temp_dir / i) for i in comms])]:
                for fn in files:  # executable files will need permissions
                    copy_file(temp_dir / fn.name, target / fn.name, permissions=0o777)

            # create links to the install directory
            linksrc = [(install_dir / i) for i in comms] + [install_dir / alphaalias]
            linkdest = [
                link_dir / "alphamelts"
                if (("alphamelts" in i.name) and (not "run" in i.name))
                else link_dir / i.name
                for i in linksrc
            ]
            if system == "Windows":  # create batch files to act as symlinks
                for src, dst in zip(linksrc, linkdest):
                    logger.debug("Creating batch file: {} <- {}".format(src, dst))
                    with open(str(dst.with_suffix(".bat")), "w") as fout:
                        fout.write("""@echo off\nperl "{}" %*""".format(src))
            else:  # create symlinks for command files and the exectuable
                for src, dst in zip(linksrc, linkdest):
                    logger.debug("Creating symlink: {} <- {}".format(src, dst))
                    if dst.exists():
                        os.remove(str(dst))  # remove old symlinks if present
                    os.symlink(str(src), str(dst))

    except AssertionError:
        raise AssertionError
    finally:
        if not keep_tempdir:
            remove_tempdir(temp_dir)



import os, sys, platform
import subprocess
import threading
import stat
import psutil
import queue
import time
from pathlib import Path
from ..util.general import get_local_link, get_process_tree
from ..util.log import Handle

logger = Handle(__name__)


def enqueue_output(out, queue):
    """
    Send output to a queue.

    Parameters
    -----------
    out
        Readable output object.
    queue : :class:`queue.Queue`
        Queue to send ouptut to.
    """
    for line in iter(out.readline, b""):
        queue.put(line)
    out.close()


class MeltsProcess(object):
    def __init__(
        self,
        executable=None,
        env="alphamelts_default_env.txt",
        meltsfile=None,
        fromdir=r"./",
        log=logger.debug,
        timeout=None,
    ):
        """
        Parameters
        ----------
        executable : :class:`str` | :class:`pathlib.Path`
            Executable to run. Enter path to the the `run_alphamelts.command `
            script. Falls back to local installation if no exectuable is specified
            and a local instllation exists.
        env : :class:`str` | :class:`pathlib.Path`
            Environment file to use.
        meltsfile : :class:`str` | :class:`pathlib.Path`
            Path to meltsfile to use for calculations.
        fromdir : :class:`str` | :class:`pathlib.Path`
            Directory to use as the working directory for the execution.
        log : :class:`callable`
            Function for logging output.

        Todo
        -----
            * Recognise errors from stdout
            * Input validation (graph of available options vs menu level)
            * Logging of failed runs
            * Facilitation of interactive mode upon error
            * Error recovery methods (e.g. change the temperature)

        Notes
        ------
            * Need to specify an exectuable or perform a local installation of alphamelts.
            * Need to get full paths for melts files, directories etc
        """
        self.env = None
        self.meltsfile = None
        self.fromdir = None  # default to None, runs from cwd
        self.log = log
        self.timeout = timeout or 60.0  # 1 minute max
        if fromdir is not None:
            self.log("Setting working directory: {}".format(fromdir))
            fromdir = Path(fromdir)
            try:
                assert fromdir.exists() and fromdir.is_dir()
            except AssertionError:
                fromdir.mkdir(parents=True)
            self.fromdir = Path(fromdir)

        if executable is None:
            # check for local install
            if platform.system() == "Windows":
                local_run = get_local_link("run_alphamelts.bat")
            else:
                local_run = get_local_link("run_alphamelts.command")

            executable = local_run
            self.log(
                "Using local executable: {} @ {}".format(
                    executable.name, executable.parent
                )
            )

        executable = Path(executable)
        self.exname = str(executable.name)
        self.executable = str(executable)
        st = os.stat(self.executable)
        assert bool(stat.S_IXUSR), "User needs execution permission."
        self.run = []

        self.run.append(self.executable)  # executable file

        self.init_args = []  # initial arguments to pass to the exec before returning
        if meltsfile is not None:
            self.log("Setting meltsfile: {}".format(meltsfile))
            self.meltsfile = Path(meltsfile)
            self.run += ["-m", str(self.meltsfile)]
            self.init_args += ["1", str(self.meltsfile)]  # enter meltsfile
        if env is not None:
            self.log("Setting environment file: {}".format(env))
            self.env = Path(env)
            self.run += ["-f", str(env)]

        self.start()  # could split this out such that processes can be prepared beforehand
        time.sleep(0.5)
        self.log("Passing Inital Variables: " + " ".join(self.init_args))
        self.write(self.init_args)

    @property
    def callstring(self):
        """Get the call string such that analyses can be reproduced manually."""
        return " ".join(["cd", str(self.fromdir), "&&"] + self.run)

    def log_output(self):
        """
        Log output to the configured logger.
        """
        self.log("\n" + self.read())

    def start(self):
        """
        Start the process.

        Returns
        --------
        :class:`subprocess.Popen`
            Melts process object.
        """
        self.started = time.time()
        self.log(
            "Starting Melts Process with: " + " ".join([self.exname] + self.run[1:])
        )
        config = dict(
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=str(self.fromdir),
            close_fds=(os.name == "posix"),
        )
        self.process = subprocess.Popen(self.run, **config)
        logger.debug("Process Started with ID {}".format(self.process.pid))
        logger.debug("Reproduce using: {}".format(self.callstring))
        # Queues and Logging
        self.q = queue.Queue()
        self.T = threading.Thread(
            target=enqueue_output, args=(self.process.stdout, self.q)
        )
        self.T.daemon = True  # kill when process dies
        self.T.start()  # start the output thread

        self.errq = queue.Queue()
        self.errT = threading.Thread(  # separate thread for error reporting
            target=enqueue_output, args=(self.process.stderr, self.errq)
        )
        self.errT.daemon = True  # kill when process dies
        self.errT.start()  # start the err output thread
        return self.process

    def read(self):
        """
        Read from the output queue.

        Returns
        ---------
        :class:`str`
            Concatenated output from the output queue.
        """
        lines = []
        while not self.q.empty():
            lines.append(self.q.get_nowait().decode())
        return "".join(lines)

    @property
    def timed_out(self):
        return (time.time() - self.started) > self.timeout

    def wait(self, step=1.0):
        """
        Wait until addtions to process.stdout stop.

        Parameters
        -----------
        step : :class:`float`
            Step in seconds at which to check the stdout queue.
        """
        while True:
            size = self.q.qsize()
            time.sleep(step)
            if self.timed_out:
                self.log(
                    "Process timed out after {:2.1f} s".format(
                        time.time() - self.started
                    )
                )
                self.terminate()
                break
            elif size == self.q.qsize():
                break

    def write(self, messages, wait=True, log=False):
        """
        Send commands to the process.

        Parameters
        -----------
        messages
            Sequence of messages/commands to send.
        wait : :class:`bool`
            Whether to wait for process.stdout to finish.
        log : :class:`bool`
            Whether to log output to the logger.
        """
        for message in messages:
            msg = (str(message).strip() + str(os.linesep)).encode("utf-8")
            self.process.stdin.write(msg)
            self.process.stdin.flush()
            if wait:
                self.wait()
            if log:
                self.log(message)
                self.log_output()

    def terminate(self):
        """
        Terminate the process.

        Notes
        -------
            * Will likely terminate as expected using the command '0' to exit.
            * Otherwise will attempt to cleanup the process.
        """
        self.alphamelts_ex = []
        try:
            for p in get_process_tree(self.process.pid):
                if "alpha" in p.name():
                    self.alphamelts_ex.append(p)
            self.write("0")
            time.sleep(0.5)
        except (ProcessLookupError, psutil.NoSuchProcess):
            logger.warning("Process terminated unexpectedly.")

        try:
            self.process.stdin.close()
            self.process.terminate()
            self.process.wait(timeout=0.2)
        except ProcessLookupError:
            logger.debug("Process terminated successfully.")

        self.cleanup()

    def cleanup(self):
        for p in self.alphamelts_ex:  # kill the children executables
            try:
                # kill the alphamelts executable which can hang
                logger.debug("Terminating {}".format(p.name()))
                p.kill()
            except (psutil.NoSuchProcess, psutil.AccessDenied) as e:
                pass



import json
import hashlib
from pyrolite.util.text import slugify
from ..util.log import Handle

logger = Handle(__name__)

__abbrv__ = {"fractionate solids": "frac", "isobaric": "isobar"}


def exp_hash(d, algorithm="sha1", length=10):
    """
    Get the hash of an experiment configuration dictionary.

    Parameters
    ----------
    d : :class:`dict`
        Configuration dictonary.
    algorithm : :class:`str`
        Name of hash algorithm to use.
    length : :class:`int`
        Length of the returned index generated from the hash.

    Returns
    --------
    :class:`str`
        Hash-based index for the configuration.

    Todo
    -------

        * Consider rounding floats to 4-5 decimal places to minimise floating point
            errors and potential differences across systems
    """

    hsh = hashlib.new(algorithm)
    # sort keys for consistency regardless of insertion order (e.g. default/cfg_grid)
    cfg = {k: d[k] for k in sorted(d.keys())}
    hsh.update(json.dumps(cfg, sort_keys=True, ensure_ascii=False).encode("utf8"))
    hex = hsh.hexdigest()
    length = length or len(hex) or -0
    return hex[:length]


def exp_name(exp):
    """
    Derive an experiment name from an experiment configuration dictionary.

    Parameters
    ------------
    exp : :class:`dict`
        Dictionary of parameters and their specific values to derive an experiment name
        from.

    Todo
    ------

        This is a subset of potential parameters, hash is used to ensure uniqueness of naming.

        To avoid path length limits, we keep only a fraction of the hash length.

        Avoid using odd symbols like @.
    """

    p0, p1, t0, t1 = "", "", "", ""
    if exp.get("Initial Pressure", None) is not None:
        p0 = "{:d}".format(int(exp["Initial Pressure"] / 1000))
    if exp.get("Final Pressure", None) is not None:
        p1 = "{:d}".format(int(exp["Final Pressure"] / 1000))
    if exp.get("Initial Temperature", None) is not None:
        t0 = "{:d}".format(int(exp["Initial Temperature"]))
    if exp.get("Final Temperature", None) is not None:
        t1 = "{:d}".format(int(exp["Final Temperature"]))
    fo2 = exp.get("Log fO2 Path", "")
    fo2d = exp.get("Log fO2 Delta", "")

    titlestr = slugify(exp.get("Title", ""))
    modestr = "".join([__abbrv__.get(m, m) for m in exp.get("modes", [""])])
    pstr = p0 + ["-" + p1, ""][p1 == ""] + ["", "kbar"][p0 + p1 != ""]
    tstr = t0 + ["-" + t1, ""][t1 == ""] + ["", "C"][t0 + t1 != ""]
    fo2str = "{}{}".format(fo2d, fo2)
    chemstr = "-".join(
        ["{}{}".format(k, v) for k, v in exp.get("modifychem", {}).items()]
    )

    suppressstr = "-".join(["no_{}".format(v) for v in exp.get("Suppress", {})])
    hashstr = "{}".format(exp_hash(exp))

    return slugify(
        "".join([titlestr, modestr, pstr, tstr, fo2str, chemstr, suppressstr, hashstr])
    )



"""
Visualisation utilities for working with alphaMELTS table outputs.

Todo
-------

    * Pull out mineral compostional trends
"""
from pyrolite.util.plot import proxy_line
from ..tables.load import import_tables
from .style import phase_color, phaseID_linestyle

from ..util.log import Handle

logger = Handle(__name__)



"""
Generating synthetic data for use with alphaMELTS.
"""
from collections import OrderedDict
from pyrolite.geochem.norm import get_reference_composition
from ..util.log import Handle

logger = Handle(__name__)


def default_data_dictionary():
    """
    Data dictionary with sufficient default values to be passed to MELTS REST services
    for testing purposes.

    Returns
    --------
    :class:`dict`
        Dictionary with some default values.
    """
    d = OrderedDict()
    d["title"] = ("TestREST",)
    d["initialize"] = {
        "SiO2": 48.68,
        "TiO2": 1.01,
        "Al2O3": 17.64,
        "Fe2O3": 0.89,
        "Cr2O3": 0.0425,
        "FeO": 7.59,
        "MnO": 0.0,
        "MgO": 9.10,
        "NiO": 0.0,
        "CoO": 0.0,
        "CaO": 12.45,
        "Na2O": 2.65,
        "K2O": 0.03,
        "P2O5": 0.08,
        "H2O": 0.20,
    }
    d["calculationMode"] = "findLiquidus"
    d["constraints"] = {"setTP": {"initialT": 1200, "initialP": 1000}}
    return d


def isobaricGaleMORBexample(T0=1300, T1=800, P0=5000, P1=5000, fO2="FMQ", title=None):
    Gale_MORB = get_reference_composition("MORB_Gale2013")
    majors = [
        "SiO2",
        "Al2O3",
        "FeO",
        "MnO",
        "MgO",
        "CaO",
        "Na2O",
        "TiO2",
        "K2O",
        "P2O5",
    ]
    MORB = Gale_MORB.comp[majors].reset_index(drop=True)
    if title is not None:
        MORB["Title"] = title
    MORB["Initial Temperature"] = T0
    MORB["Final Temperature"] = T1
    MORB["Initial Pressure"] = P0
    MORB["Final Pressure"] = P1
    MORB["Log fO2 Path"] = fO2
    MORB["Increment Temperature"] = -5
    MORB["Increment Pressure"] = 0
    return MORB



"""
alphaMELTS environment managment.
"""
import os
import logging
from pyrolite.util.env import validate_update_envvar
from pyrolite.util.text import remove_prefix
from textwrap import dedent
from .data.environment import MELTS_environment_variables

logging.getLogger(__name__).addHandler(logging.NullHandler())
logger = logging.getLogger(__name__)


def output_formatter(value):
    """
    Output formatter for environment variable values.

    Parameters
    ------------
    value
        Value to format.

    Returns
    --------
    :class:`str`
        Formatted value.
    """
    if value is not None and not isinstance(value, bool):
        return str(value)
    else:
        return ""


class MELTS_Env(object):
    """
    Melts environment object.

    Todo
    -----
        * Implement use as context manager.
    """

    def __init__(self, prefix="ALPHAMELTS_", variable_model=None):
        self.prefix = prefix
        if variable_model is None:
            variable_model = MELTS_environment_variables
        self.spec = variable_model
        self.force_active = False
        self.output_formatter = output_formatter
        self.export_default_env()

    def export_default_env(self):
        """
        Parse any environment variables which are already set.
        Reset environment variables after substituting defaults for unset
        variables.
        """
        _dump = self.dump(prefix=False)

        for var, template in self.spec.items():
            _spec = template
            name = self.prefix + var
            is_already_set = (name in os.environ) or (var in _dump.keys())
            if not is_already_set and _spec["set"]:
                setting = True  # should be set by default
            elif is_already_set and _spec["set"]:
                setting = True  # set, should be set by default
            elif is_already_set and not _spec["set"]:
                setting = True  # set, not set by default
            elif not is_already_set and not _spec["set"]:
                setting = False  # not set, should not be set

            if setting:
                setattr(self, var, None)

    def dump(self, unset_variables=True, prefix=False, cast=lambda x: x):
        r"""
        Export environment configuration to a dictionary.

        Parameters
        -----------
        unset_variables : :class:`bool`
            Whether to include variables which are currently unset.
        prefix : :class:`bool`
            Whether to prefix environment variables (i.e with ALPHAMELTS\_).
        cast : :class:`callable`
            Function to cast environment variable values.

        Returns
        --------
        :class:`dict`
            Dictionary of environent variables and their values.
        """
        keys = [k for k in self.spec.keys()]
        pkeys = [self.prefix + k for k in keys]
        values = [os.getenv(p) for p in pkeys]
        types = [
            self.spec[k]["type"] if self.spec[k].get("type", None) is not None else str
            for k in keys
        ]

        # Evironment variable are always imported as strings
        _env = [
            (k, t(v)) if v and (v not in [None, "None"]) else (k, None)
            for k, p, v, t in zip(keys, pkeys, values, types)
        ]
        if not unset_variables:
            _env = [e for e in _env if e[1] is not None]
        return {[k, self.prefix + k][prefix]: cast(v) for k, v in _env}

    def to_envfile(self, unset_variables=False):
        """
        Create a string representation equivalent to the alphamelts defualt
        environment file.

        Parameters
        -----------
        unset_variables : :class:`bool`
            Whether to include unset variables in the output (commented out).

        Returns
        -------
        :class:`str`
            String-representation of the environment which can be writen to a file.
        """
        preamble = dedent(
            """
        ! Default values of environment variables (pyrolite export)
        ! Variables preceeded by '!' are 'unset' (i.e. 'false')
        """
        )
        return preamble + "\n".join(
            [
                ["", "!"][v is None] + "{} {}".format(k, v)
                for k, v in self.dump(
                    prefix=True, unset_variables=unset_variables
                ).items()
            ]
        )

    def __setattr__(self, name, value):
        """
        Custom setattr to set environment variables.

        Setting attributes with or without the specified prefix should set
        the appropriate prefixed environment variable.
        """

        if hasattr(self, "spec"):
            prefix = getattr(self, "prefix", "")
            dump = self.dump()
            name = remove_prefix(name, prefix)
            if name in self.spec:
                validate_update_envvar(
                    name,
                    value=value,
                    prefix=self.prefix,
                    force_active=self.force_active,
                    variable_model=self.spec,
                    formatter=self.output_formatter,
                )
            else:  # other object attributes
                self.__dict__[name] = value
        else:
            self.__dict__[name] = value

    def __repr__(self):
        """Returns the class and all set variables."""
        return "{}({})".format(
            self.__class__.__name__, self.dump(unset_variables=False)
        ).replace(",", ",\n\t\t")



from ..util.log import Handle

logger = Handle(__name__)

def get_phaselist(folder):
    """
    Get the list of phases from an experiment.

    Parameters
    -----------
    folder : :class:`str` | :class:`pathlib.Path`
        Path to the experiment directory.
    """


def get_assemblages(folder):
    """
    Get the list of phases from an experiment.

    Parameters
    -----------
    folder : :class:`str` | :class:`pathlib.Path`
        Path to the experiment directory.
    """



from ..util.tables import phasename
import matplotlib.colors as mcolors
from ..util.log import Handle

logger = Handle(__name__)

COLORS = {
    "aegirine": None,
    "aenigmatite": None,
    "alloy-liquid": None,
    "alloy-solid": None,
    "amphibole": "darkslategrey",
    "apatite": None,
    "biotite": "brown",
    "clinopyroxene": "teal",
    "corundum": None,
    "cristobalite": None,
    "cummingtonite": None,
    "fayalite": None,
    "feldspar": "pink",
    "garnet": "red",
    "hornblende": "darkolivegreen",
    "kalsilite": None,
    "liquid": "black",
    "leucite": None,
    "melilite": None,
    "muscovite": None,
    "nepheline": None,
    "olivine": "green",
    "ortho-oxide": None,
    "orthopyroxene": "darkorange",
    "perovskite": None,
    "quartz": None,
    "rhm-oxide": None,
    "rutile": None,
    "sillimanite": None,
    "sphene": None,
    "spinel": "0.5",
    "tridymite": None,
    "water": "aquamarine",
    "whitlockite": None,
}


def phase_color(phase, rgb=False):
    """
    Method for generating colors for delineating phase names
    (e.g. olivine, clinopyroxene) based on their names.

    Parameters
    ------------
    phase : :class:`str`
        Phase name or phase ID to generate a color for.

    Returns
    --------
    :class:`str`
        Color for the phase.
    """
    c = COLORS.get(phasename(phase), None)
    if rgb:
        c = mcolors.to_rgb(c)  # will error on None
    return c


def phaseID_linestyle(phaseID, linestyles=["-", "--", ":", "-."]*2):
    """
    Method for generating linestyles for delineating sequential phases
    based on their phase IDs (e.g. olivine_0, olivine_1) .

    Parameters
    -----------
    phasename : :class:`str`
        Phase ID for which to generate a line style.
    linestyles : :class:`list`
        List of line styles for sequential phase ID numbers.
        Added to allow reconfiguraiton where needed.

    Returns
    ---------
    :class:`str`
        Line style for the phase ID.
    """
    if "_" in phaseID:
        return linestyles[int(phaseID[-1])]
    else:
        return linestyles[0]


def phaseID_marker(phaseID, markers=["D", "s", "o", "+", "*"]*2):
    """
    Method for generating markers for delineating sequential phases
    based on their phase IDs (e.g. olivine_0, olivine_1) .

    Parameters
    -----------
    phasename : :class:`str`
        Phase ID for which to generate a line style.
    markers : :class:`list`
        List of markers for sequential phase ID numbers.
        Added to allow reconfiguraiton where needed.

    Returns
    ---------
    :class:`str`
        Marker for the phase ID.
    """
    if "_" in phaseID:
        return markers[int(phaseID[-1])]
    else:
        return markers[0]



"""
This submodule is a home for templates for quickly
regenerating common visualistions from melts tables.
"""
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from pyrolite.util.meta import subkwargs
from pyrolite.util.plot import proxy_line
from .style import phase_color, phaseID_linestyle, phaseID_marker
from ..util.log import Handle

logger = Handle(__name__)


def plot_xy_phase_groupby(ax, df, xvar, yvar, legend=True, markersize=3, **kwargs):
    """
    Utility function for plotting phases on an axis.

    Returns
    ---------
    :class:`matplotlib.axes.Axes`
        Axes on which the table results are plotted.
    """

    if "experiment" in df.columns:
        assert len(df.experiment.unique()) <= 1
    if ax is None:
        fig, ax = plt.subplots(1)
    phaseIDlist = sorted([i for i in df.phaseID.unique() if not pd.isnull(i)])
    ax.set_xlabel(xvar)
    ax.set_ylabel(yvar)
    proxies = {}

    xinds = df.step.drop_duplicates().sort_values().index  # steps are unique

    for phaseID in phaseIDlist:
        phasedf = df.loc[df.phaseID == phaseID, :].reindex(index=xinds)
        style = dict(
            ls=phaseID_linestyle(phaseID),
            color=phase_color(phaseID),
            markerfacecolor=phase_color(phaseID),
            marker=phaseID_marker(phaseID),
            markersize=markersize,
        )
        style = {**{k: v for k, v in style.items() if v is not None}, **kwargs}
        proxies[phaseID] = proxy_line(**style)
        ax.plot(phasedf[xvar].values, phasedf[yvar].values, **style)

    if legend:
        ax.legend(
            proxies.values(),
            proxies.keys(),
            frameon=False,
            facecolor=None,
            loc="upper left",
            bbox_to_anchor=(1, 1),
        )
    return ax, proxies


def plot_phasevolumes(phasetable, xvar="temperature", legend=True, ax=None, **kwargs):
    """
    """
    ax, proxies = plot_xy_phase_groupby(
        ax, phasetable, xvar, "volume%", legend=legend, **kwargs
    )
    return ax, proxies


def plot_phasemasses(phasetable, xvar="temperature", legend=True, ax=None, **kwargs):
    """
    """
    ax, proxies = plot_xy_phase_groupby(
        ax, phasetable, xvar, "mass%", legend=legend, **kwargs
    )
    return ax, proxies



import psutil
import subprocess
from pyrolite.util.meta import get_module_datafolder
from ..util.log import Handle

logger = Handle(__name__)

def check_perl():
    """
    Checks whether perl is installed on the system.

    Returns
    -------
    :class:`bool`
        Boolean indication of whether there is an executable perl installation.
    """
    try:
        p = subprocess.check_output(["perl", "-v"])
        returncode = 0
    except subprocess.CalledProcessError as e:
        output = e.output
        returncode = e.returncode
    except FileNotFoundError:
        returncode = 1

    return returncode == 0

def get_process_tree(process, levels_up=1):
    """
    Get a process tree from an active process or process ID.

    Parameters
    -----------
    process : :class:`int` | :class:`psutil.Process`
        Process to search for.
    levels_up : :class:`int`
        How many levels up the tree to search for parent processes.

    Returns
    -------
    :class:`list`
        List of processes associated with the given process tree.
    """
    if isinstance(process, int):
        top = psutil.Process(process)
    elif isinstance(process, psutil.Process):
        top = process
    for i in range(levels_up):
        if top.parent() is not None:
            top = top.parent()
        else:
            break
    return [top, *top.children(recursive=True)]


def pyrolite_meltsutil_datafolder(subfolder=None):
    """
    Returns the path of the pyrolite-meltsutil data folder.

    Parameters
    -----------
    subfolder : :class:`str`
        Subfolder within the pyrolite data folder.

    Returns
    -------
    :class:`pathlib.Path`
    """
    return get_module_datafolder(module="pyrolite_meltsutil", subfolder=subfolder)


def get_local_link(name):
    """
    Get the filepath of a local link file installed with alphaMELTS.

    Parameters
    -----------
    name : :class:`str`
        Filename for the link file.

    Returns
    -------
    :class:`pathlib.Path`
    """
    return pyrolite_meltsutil_datafolder("localinstall") / "links" / name


def get_local_example(name):
    """
    Get the filepath of a local example file installed with alphaMELTS.

    Parameters
    -----------
    name : :class:`str`
        Filename for the example file.

    Returns
    -------
    :class:`pathlib.Path`
    """
    return pyrolite_meltsutil_datafolder("localinstall") / "examples" / name


def get_data_example(name):
    """
    Get the filepath of an example data file from pyrolite-meltsutil.

    Parameters
    -----------
    name : :class:`str`
        Filename for the example file.

    Returns
    -------
    :class:`pathlib.Path`
    """
    return pyrolite_meltsutil_datafolder("data_examples") / name



from ..util.log import Handle

logger = Handle(__name__)

TIME_PER_STEP = 0.05  # seconds per iteration, varies


def get_P_steps(exp):
    """
    """

    if "isobaric" in exp.get("modes", []):
        steps = 0
    else:
        steps = 1


def estimate_experiment_duration(experiment):
    n, exp, env = experiment
    # get pressure steps

    # get temperature steps

    # check if isobaric or isothermal

    #



from setuptools import setup, find_packages
import versioneer

tests_require = ["pytest", "pytest-runner", "pytest-cov", "coverage", "coveralls"]

dev_require = [
    "pytest",
    "versioneer",
    "black",
    "twine",
    "sphinx_rtd_theme",
    "sphinx-autodoc-annotation",
    "sphinx_gallery",
    "recommonmark",
] + tests_require

with open("README.md", "r") as src:
    LONG_DESCRIPTION = src.read()

setup(
    name="pyrolite-meltsutil",
    description="pyrolite extension for working with alphaMELTS and its outputs.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    version=versioneer.get_version(),
    url="https://github.com/morganjwilliams/pyrolite-meltsutil",
    project_urls={
        "Documentation": "https://pyrolite-meltsutil.readthedocs.com/",
        "Code": "https://github.com/morganjwilliams/pyrolite-meltsutil",
        "Issue tracker": "https://github.com/morganjwilliams/pyrolite-meltsutil/issues",
    },
    author="Morgan Williams",
    author_email="morgan.williams@csiro.au",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords=["geochemistry", "compositional data", "visualisation", "petrology"],
    packages=find_packages(exclude=["test*"]),
    install_requires=[
        "pyrolite>=0.2.2",
        "requests",
        "psutil",
        "xmljson",
        "dicttoxml",
        "tqdm",
    ],
    extras_require={"dev": dev_require},
    tests_require=tests_require,
    test_suite="test",
    include_package_data=True,
    license="CSIRO Modifed MIT/BSD",
    cmdclass=versioneer.get_cmdclass(),
)



"""
This submodule contains functions for automated execution of alphamelts 1.9.

Issues
------

    * names are truncated for modifychem melts files?
    * need a timeout so processes can keep going, add unfinished experiments to failed list
"""
import itertools
from pathlib import Path
import time, datetime
import numpy as np
import json
from tqdm import tqdm

from pyrolite.geochem.ind import common_elements, common_oxides
from pyrolite.comp.codata import renormalise
from pyrolite.util.meta import ToLogger
from pyrolite.util.multip import combine_choices

from ..parse import read_envfile, read_meltsfile
from ..env import MELTS_Env
from ..meltsfile import dict_to_meltsfile

from .naming import exp_name, exp_hash
from .org import make_meltsfolder
from .process import MeltsProcess
from .timing import estimate_experiment_duration

import logging
from ..util.log import Handle

logger = Handle(__name__)

__chem__ = common_elements(as_set=True) | common_oxides(as_set=True)


class MeltsExperiment(object):
    """
    Melts Experiment Object. For a single call to melts, with one set of outputs.
    Autmatically creates the experiment folder, meltsfile and environment file, runs
    alphaMELTS and collects the results.

    Todo
    ----
        * Automated creation of folders for experiment results (see :func:`make_meltsfolder`)
        * Being able to run melts in an automated way (see :class:`MeltsProcess`)
        * Compressed export/save function
        * Post-processing functions for i) validation and ii) plotting
    """

    def __init__(
        self,
        name="MeltsExperiment",
        title="MeltsExperiment",
        fromdir="./",
        meltsfile=None,
        env=None,
        timeout=None,
    ):
        self.name = name  # folder name
        self.title = title  # meltsfile title
        self.fromdir = fromdir  # create an experiment directory here
        self.log = []
        self.timeout = timeout

        if meltsfile is not None:
            self.set_meltsfile(meltsfile)
        if env is not None:
            self.set_envfile(env)
        else:
            self.set_envfile(MELTS_Env())

        self._make_folder()

    def set_meltsfile(self, meltsfile, **kwargs):
        """
        Set the meltsfile for the experiment.

        Parameters
        ------------
        meltsfile : :class:`pandas.Series` | :class:`str` | :class:`pathlib.Path`
            Either a path to a valid melts file, a :class:`pandas.Series`, or a
            multiline string representation of a melts file object.
        """
        self.meltsfile, self.meltsfilepath = read_meltsfile(meltsfile)

    def set_envfile(self, env):
        """
        Set the environment for the experiment.

        Parameters
        ------------
        env : :class:`str` | :class:`pathlib.Path`
            Either a path to a valid environment file, a :class:`pandas.Series`, or a
            multiline string representation of a environment file object.
        """
        self.envfile, self.envfilepath = read_envfile(env)

    def _make_folder(self):
        """
        Create the experiment folder.
        """
        self.folder = make_meltsfolder(
            name=self.name,
            title=self.title,
            meltsfile=self.meltsfile,
            indir=self.fromdir,
            env=self.envfile,
        )
        self.meltsfilepath = self.folder / (self.title + ".melts")
        self.envfilepath = self.folder / "environment.txt"

    def run(self, log=False, superliquidus_start=True):
        """
        Call 'run_alphamelts.command'.
        """
        self.mp = MeltsProcess(
            meltsfile=str(self.title) + ".melts",
            env="environment.txt",
            fromdir=str(self.folder),
            timeout=self.timeout,
        )
        self.mp.write([3, [0, 1][superliquidus_start], 4], wait=True, log=log)
        self.mp.terminate()

    def cleanup(self):
        pass


def process_modifications(cfg):
    """
    Process modifications to an configuration composition.

    Parameters
    -----------
    cfg : :class:`dict`
        Configuratiion dictionary.

    Returns
    --------
    cfg : :class:`dict`
        Configuratiion dictionary.
    """
    if "modifychem" in cfg:
        modifications = cfg.pop("modifychem", {})  # remove modify chem
        ek, mk = set(cfg.keys()), set(modifications.keys())
        for k, v in modifications.items():
            if not np.isnan(v):
                cfg[k] = v
        allchem = (ek | mk) & __chem__
        unmodified = (ek - mk) & __chem__

        offset = np.nansum(np.array(list(modifications.values())))
        for uk in unmodified:
            cfg[uk] = np.round(cfg[uk] * (100.0 - offset) / 100, 4)
    return cfg


class MeltsBatch(object):
    """
    Batch of :class:`MeltsExperiment`, which may represent evaluation over a grid of
    parameters or configurations.

    Parameters
    -----------
    comp_df : :class:`pandas.DataFrame`
        Dataframe of compositions.
    default_config : :class:`dict`
        Dictionary of default parameters.
    config_grid : class:`dict`
        Dictionary of parameters to systematically vary.

    Attributes
    -----------

    compositions : :class:`list` of :class:`dict`
        Compositions to use for

    configs : :class:`list` of :class:`dict`

    experiments : :class:`list` of :class:`dict`

    Todo
    ------
        * Can start with a single composition or multiple compositions in a dataframe
        * Enable grid search for individual parameters
        * Improved output logging/reporting
        * Calculate relative number of calculations to be performed for the est duration

            This is currently about correct for an isobaric calcuation at 10 degree
            temperature steps over few hundred degrees - but won't work for different
            T steps.

        * Does number precision make a difference?
    """

    def __init__(
        self,
        comp_df,
        fromdir=Path("./"),
        default_config={},
        config_grid={},
        env=None,
        logger=logger,
        timeout=None,
    ):
        self.timeout = timeout
        self.logger = logger
        self.fromdir = Path(fromdir)
        # make a file logger
        fh = logging.FileHandler(self.fromdir / "autolog.log")
        fh.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        fh.setFormatter(formatter)
        self.logger.addHandler(fh)

        self.default = default_config
        self.env = env or MELTS_Env()
        # let's establish the grid of configurations
        self.configs = []
        grid = combine_choices(config_grid)
        for i in grid:  # unique configurations
            _cfg = {**self.default, **i}
            if _cfg not in self.configs:
                self.configs.append(_cfg)
        self.compositions = comp_df.fillna(0).to_dict("records")
        # combine these to create full experiment configs
        exprs = [
            {**cfg, **cmp}
            for (cfg, cmp) in itertools.product(self.configs, self.compositions)
        ]
        exprs = [process_modifications(cfg) for cfg in exprs]
        exphashes = np.array([exp_hash(i) for i in exprs])
        _, cnts = np.unique(exphashes, return_counts=True)
        if (cnts > 1).any():
            self.logger.debug("Duplicate experiments detected.")
        self.experiments = {
            hsh: (exp_name(expr), expr, self.env) for hsh, expr in zip(exphashes, exprs)
        }  # this ensures that no duplicates are preserved

        self.est_duration = str(
            datetime.timedelta(seconds=len(self.experiments) * 15)
        )  # 6s/run
        self.logger.info("Estimated Calculation Time: {}".format(self.est_duration))

    def dump(self, experiments=None, to_dir=None):
        """
        Serialize the configuration to a json file.

        Parameters
        -----------
        experiments : :class:`dict`
            Dictionary of experiments to be serialized.
        to_dir : :class:`str` | :class:`pathlib.Path`
            Directory to export file to.
        """
        to_dir = to_dir or self.fromdir
        experiments = experiments or self.experiments
        data = json.dumps(
            {
                h: (t, exp, env.dump(unset_variables=False))
                for (h, (t, exp, env)) in experiments.items()
            },
            sort_keys=False,
            ensure_ascii=False,
        ).encode("utf8")

        target = Path(to_dir) / "meltsBatchConfig.json"
        target.parent.mkdir(parents=True, exist_ok=True)  # may not exist yet?
        # consider reading old data and leaving updated version here
        target.touch(exist_ok=True)
        with open(target, "wb") as f:
            f.write(data)

    def run(self, overwrite=False, exclude=[], superliquidus_start=True, timeout=None):
        self.dump()  # Serialize the config first
        timeout = self.timeout or timeout
        self.started = time.time()
        experiments = self.experiments
        if not overwrite:
            experiments = {
                h: (t, exp, env)
                for h, (t, exp, env) in experiments.items()
                if not (self.fromdir / h).exists()
            }

        self.logger.info("Starting {} Calculations.".format(len(experiments)))
        paths = []
        failed = []
        for hsh, (title, exp, env) in tqdm(
            experiments.items(), file=ToLogger(self.logger), mininterval=2
        ):
            exp_exclude = exclude
            if "exclude" in exp:
                exp_exclude += exp.pop("exclude")  # remove exclude

            self.logger.debug("Start {}.".format(title))
            meltsfile = dict_to_meltsfile(exp, modes=exp["modes"], exclude=exp_exclude)
            M = MeltsExperiment(
                name=hsh,
                title=title,
                meltsfile=meltsfile,
                env=env,
                fromdir=self.fromdir,
                timeout=timeout,
            )
            try:
                M.run(superliquidus_start=superliquidus_start)
                self.logger.debug("Finished {}.".format(title))
            except OSError:
                try:
                    self.logger.warning("Errored @ {}.".format(M.mp.callstring))
                except:
                    pass
                failed.append(title)
        # should check if it actually ran here (e.g. timeouts)
        self.duration = datetime.timedelta(seconds=time.time() - self.started)
        self.logger.info("Calculations Complete after {}".format(self.duration))
        if failed:
            self.logger.warning("Some calculations errored:")
            for f in failed:
                self.logger.warning(f)

    def cleanup(self):
        pass



"""
Parsing utilities for use with alphaMELTS.
"""
import re
import pandas as pd
from pathlib import Path
import logging
import periodictable as pt
from pyrolite.mineral.transform import merge_formulae
from .env import MELTS_Env
from .meltsfile import ser_to_meltsfile

logging.getLogger(__name__).addHandler(logging.NullHandler())
logger = logging.getLogger(__name__)


def _file_from_obj(fileobj):
    """
    Read in file data either from a file path or a string.

    Parameters
    ------------
    fileobj : :class:`str` | :class:`pathlib.Path`
        Either a path to a valid file, or a multiline string representation of a
        file object.

    Returns
    --------
    file : :class:`str`
        Multiline string representation of a file.
    path
        Path to the original file, if it exists.

    Notes
    ------
        This function deconvolutes the possible ways in which one can pass either
        a file, or reference to a file.

    Todo
    ----
        * Could be passed an open file object
    """
    path, file = None, None
    if isinstance(fileobj, Path):
        path = str(fileobj)
    elif isinstance(fileobj, str):
        if len(re.split("[\r\n]", fileobj)) > 1:  # multiline string passed as a file
            file = fileobj
        else:  # path passed as a string
            path = fileobj
    else:
        pass

    if (path is not None) and (file is None):
        with open(path) as f:
            file = f.read()

    assert file is not None  # can't not have a meltsfile
    return file, path


def read_meltsfile(meltsfile, **kwargs):
    """
    Read in a melts file from a :class:`~pandas.Series`, :class:`~pathlib.Path` or
    string.

    Parameters
    ------------
    meltsfile : :class:`pandas.Series` | :class:`str` | :class:`pathlib.Path`
        Either a path to a valid melts file, a :class:`pandas.Series`, or a
        multiline string representation of a melts file object.

    Returns
    --------
    file : :class:`str`
        Multiline string representation of a meltsfile.
    path
        Path to the original file, if it exists.

    Notes
    ------
        This function deconvolutes the possible ways in which one can pass either
        a meltsfile, or reference to a meltsfile.
    """
    path, file = None, None
    if isinstance(meltsfile, pd.Series):
        file = to_meltsfile(meltsfile, **kwargs)
    else:
        file, path = _file_from_obj(meltsfile)
    return file, path


def read_envfile(envfile, **kwargs):
    """
    Read in a environment file from a  :class:`~pyrolite.util.alphamelts.env.MELTS_Env`,
    :class:`~pathlib.Path` or string.

    Parameters
    ------------
    envfile : :class:`~pyrolite.util.alphamelts.env.MELTS_Env` | :class:`str` | :class:`pathlib.Path`
        Either a path to a valid environment file, a :class:`pandas.Series`, or a
        multiline string representation of a environment file object.

    Returns
    --------
    file : :class:`str`
        Multiline string representation of an environment file.
    path
        Path to the original file, if it exists.
    """
    path, file = None, None
    if isinstance(envfile, MELTS_Env):
        file = envfile.to_envfile(**kwargs)
    else:
        file, path = _file_from_obj(envfile)
    return file, path


def from_melts_cstr(composition_str, formula=True):
    """
    Parses melts composition strings to dictionaries or formulae.

    Parameters
    -----------
    composition_str : :class:`str`
        Composition to parse.
    formula : :class:`bool`
        Whether to output a :class:`periodictable.formula.Formula`

    Returns
    --------
    :class:`dict` | :class:`periodictable.formulas.Formula`
        Dictionary containing components, or alternatively if :code:`formula = True`,
        a :class:`~periodictable.formulas.Formula` representation of the composition.

    Todo
    ------
        * Enable parsing of nested brackets in composition.
    """
    regex = r"""(?P<el>[a-zA-Z']+)(?P<num>[^a-zA-Z()]+)"""
    sub = r"""[\']+"""

    def repl(m):
        return (
            "{" + str(m.group(0).count("""'""")) + "+" + "}"
        )  # replace ' with count(')

    # replace iron valences
    tfm = lambda s: re.sub(sub, repl, s).replace('[]', '')
    if not formula:
        result = re.findall(regex, composition_str)
        result = [(tfm(el), float(val)) for el, val in result]
        return {k: v for k, v in result}
    else:
        # entries with rounding errors giving negative 0 won't match
        result = tfm(str(composition_str).replace("nan", "").replace("-0.0", "0.0"))
        return pt.formula(result)



"""
Minimal REST interfaces to the MELTS web services.

Todo
------

    * Function to generate a valid parameter dictionary to pass to the service.
"""
import logging
import requests
import dicttoxml, xmljson
from xml.etree import ElementTree as ET
from pyrolite.util.web import internet_connection

logging.getLogger(__name__).addHandler(logging.NullHandler())
logger = logging.getLogger(__name__)


def melts_query(data_dict, url_sfx="Compute"):
    """
    Execute query against the MELTS web services.

    Parameters
    ----------
    data_dict : :class:`dict`
        Dictionary containing data to be sent to the web query.
    url_sfx : :class:`str`, :code:`Compute`
        URL suffix to denote specific web service (Compute | Oxides | Phases).

    Returns
    --------
    :class:`dict`
        Dictionary containing results.
    """
    try:
        assert internet_connection()
        url = "http://thermofit.ofm-research.org:8080/multiMELTSWSBxApp/" + url_sfx
        xmldata = dicttoxml.dicttoxml(
            data_dict, custom_root="MELTSinput", root=True, attr_type=False
        )
        headers = {"content-type": "text/xml", "data-type": "xml"}
        resp = requests.post(url, data=xmldata, headers=headers)
        resp.raise_for_status()
        result = xmljson.parker.data(ET.fromstring(resp.text))
        return result
    except AssertionError:
        raise AssertionError("Must be connected to the internet to run query.")


def melts_compute(data_dict):
    """
    Execute 'Compute' query against the MELTS web services.

    Parameters
    ----------
    data_dict : :class:`dict`
        Dictionary containing data to be sent to the Compute web query.

    Returns
    --------
    :class:`dict`
        Dictionary containing results.
    """
    url_sfx = "Compute"
    result = melts_query(data_dict, url_sfx=url_sfx)
    assert "Success" in result["status"]
    return result


def melts_oxides(data_dict):
    """
    Execute 'Oxides' query against the MELTS web services.

    Parameters
    ----------
    data_dict : :class:`dict`
        Dictionary containing data to be sent to the Oxides web query.

    Returns
    --------
    :class:`dict`
        Dictionary containing results.
    """
    model = data_dict["initialize"].pop("modelSelection", "MELTS_v1.0.x")
    data_dict = {"modelSelection": model}
    url_sfx = "Oxides"
    result = melts_query(data_dict, url_sfx=url_sfx)
    return result["Oxide"]


def melts_phases(data_dict):
    """
    Execute 'Phases' query against the MELTS web services.

    Parameters
    ----------
    data_dict : :class:`dict`
        Dictionary containing data to be sent to the Phases web query.

    Returns
    --------
    :class:`dict`
        Dictionary containing results.
    """
    model = data_dict["initialize"].pop("modelSelection", "MELTS_v1.0.x")
    data_dict = {"modelSelection": model}
    url_sfx = "Phases"
    result = melts_query(data_dict, url_sfx=url_sfx)
    return result["Phase"]



import numpy as np
import matplotlib.pyplot as plt
import pyrolite_meltsutil as melts
from pyrolite.util.plot import get_twins, share_axes
from pyrolite.geochem.magma import SCSS
from ..util.log import Handle

logger = Handle(__name__)

def plot_sulfur_saturation_point(
    liquid, ax=None, start=1000, xvar="mass%", mode="sulfide"
):
    """
    Plot a perfectly incompatbile trend for sulfur in a melt, and
    indicate the point at which it crosses a previously-calculated
    sulfur saturation point.

    Parameters
    -------------
    liquid : :class:`pandas.DataFrame`
        Liquid composition table.
    ax : :class:`matplotlib.axes.Axes`
        Optionally specify an axis to plot on.
    start : :class:`float` | :class:`list`
        Starting abundance of sulfur in ppm.
    xvar : :class:`str`
        X-variable for the plot.
    mode : :class:`str`
        Where SCSS is missing, add the value from either the `sulfide` saturation or
        `sulfate` saturation model.

    Returns
    -------
    ax : :class:`matplotlib.axes.Axes`
        Axes with saturation points indicated.
    """
    if ax is None:
        fig, ax = plt.subplots(1)
    ax2 = get_twins(ax)
    if not ax2:
        ax2 = ax.twinx()
    else:
        ax2 = ax2[0]
    ax2.set_ylabel("Free S (Mass %)")

    if not "SCSS" in liquid.columns: # add SCSS if its missing
        liquid["SCSS"] = SCSS(
            liquid,
            T=liquid.temperature,
            P=liquid.pressure / 1000,
            grid=None,
            kelvin=False,
        )[mode == "sulfide"]
    # plot the SCSS
    ax.plot(liquid[xvar], liquid["SCSS"], color="k", zorder=-1)
    # plot the sulfur abundance in the melt
    # this will follow SCSS after sulfide saturation
    if isinstance(start, (int, float)):
        start = np.array([start])
    elif isinstance(start, (list, tuple, np.ndarray)):
        start = np.array(start)

    saturation_info = {}
    for xS in start:
        s_wtpct = xS / 10000.0
        s_wtpct /= liquid["mass%"] / liquid["mass%"].values[0]
        satpoint = np.argmax(
            s_wtpct.values > liquid["SCSS"].values
        )  # will return where first true
        satabund = s_wtpct[satpoint]

        s_melt = np.ones_like(s_wtpct) * np.nan
        s_melt[s_wtpct < liquid.SCSS] = s_wtpct[s_wtpct < liquid.SCSS]
        s_melt[s_wtpct >= liquid.SCSS] = liquid.SCSS[s_wtpct >= liquid.SCSS]
        lines = ax.plot(
            liquid[xvar], s_melt, ls="--", label="Sulfur Content ({} ppm)".format(start)
        )
        # plot the excess sulfur abundnace
        s_free = np.ones_like(s_wtpct) * np.nan
        s_free[s_wtpct >= liquid.SCSS] = (
            s_wtpct[s_wtpct >= liquid.SCSS] - liquid.SCSS[s_wtpct >= liquid.SCSS]
        )
        ax2.plot(liquid[xvar], s_free, color=lines[0].get_color())
        saturation_info[xS] = dict(x=liquid[xvar][satpoint], y=s_wtpct[satpoint])

    x0, y0 = liquid[xvar].values[0], liquid["SCSS"].values[0]
    starts_saturated = [k for (k, d) in saturation_info.items() if d["x"] == x0]
    if starts_saturated:
        ax.annotate(
            "Start {} ppm".format(
                "({})".format(",".join([str(s) for s in starts_saturated]))
            ),
            xytext=(
                x0 + (0.05 * (np.nanmax(liquid[xvar]) - np.nanmin(liquid[xvar]))),
                0.5 * y0,
            ),
            xy=(x0, y0),
            rotation=90,
            fontsize=10,
            ha="right",
            va="center",
            arrowprops=dict(shrink=0.05, width=1, headwidth=1),
        )
    for k, d in saturation_info.items():  # annotate the (first) saturation point
        if k not in starts_saturated:
            ax.annotate(
                "{:.4g} @ {} ppm".format(d["x"], k),
                xytext=(d["x"], 1.2 * d["y"]),
                xy=(d["x"], d["y"]),
                rotation=90,
                fontsize=10,
                ha="center",
                va="bottom",
                arrowprops=dict(shrink=0.05, width=1, headwidth=1),
            )
    share_axes([ax, ax2], which="xy")
    return ax



"""
Utilities for working with the alphaMELTS
executable and associated tabular data. Note that these are
currently experimental and not affiliated with alphaMELTS. In the future, these
utilities will likely also make use of the under-development `python-melts`.

Todo
-----
     * As it is developed, also make use of python-melts.
     * Develop functions for automation over a grid (of P, T, H2O, fO2, X)
     * Have an option to aggregate summary data, and options to discard experiment data
     * Expansion of documentation

See the `alphaMELTS site <https://magmasource.caltech.edu/alphamelts/>`__ for more info
[#ref_1]_  [#ref_2]_  [#ref_3]_  [#ref_4]_ [#ref_5]_ [#ref_6]_ [#ref_7]_.

References
-----------
.. [#ref_1] Ghiorso M. S. and Sack R. O. (1995). Chemical mass transfer in magmatic
        processes IV. A revised and internally consistent thermodynamic model for the
        interpolation and extrapolation of liquid-solid equilibria in magmatic systems
        at elevated temperatures and pressures. Contributions to Mineralogy and
        Petrology 119, 197–212.
        doi: {ghiorso1995}
.. [#ref_2] Ghiorso M. S., Hirschmann M. M., Reiners P. W. and Kress V. C. (2002).
        The pMELTS: A revision of MELTS for improved calculation of phase relations
        and major element partitioning related to partial melting of the mantle to
        3 GPa. Geochemistry, Geophysics, Geosystems 3, 1–35.
        doi: {ghiorso2002}
.. [#ref_3] Asimow P. D., Dixon J. E. and Langmuir C. H. (2004).
        A hydrous melting and fractionation model for mid-ocean ridge basalts:
        Application to the Mid-Atlantic Ridge near the Azores. Geochemistry,
        Geophysics, Geosystems 5.
        doi: {asimow2004}
.. [#ref_4] Smith P. M. and Asimow P. D. (2005).
        Adiabat_1ph: A new public front-end to the MELTS, pMELTS, and pHMELTS models.
        Geochemistry, Geophysics, Geosystems 6.
        doi: {smith2005}
.. [#ref_5] Thompson R. N., Riches A. J. V., Antoshechkina P. M., Pearson D. G.,
        Nowell G. M., Ottley C. J., Dickin A. P., Hards V. L., Nguno A.-K. and
        Niku-Paavola V. (2007). Origin of CFB Magmatism: Multi-tiered Intracrustal
        Picrite–Rhyolite Magmatic Plumbing at Spitzkoppe, Western Namibia, during
        Early Cretaceous Etendeka Magmatism. J Petrology 48, 1119–1154.
        doi: {thompson2007}
.. [#ref_6] Antoshechkina P. M., Asimow P. D., Hauri E. H. and Luffi P. I. (2010).
        Effect of water on mantle melting and magma differentiation, as modeled using
        Adiabat_1ph 3.0. AGU Fall Meeting Abstracts 53, V53C-2264.
.. [#ref_7] Antoshechkina P. M. and Asimow P. D. (2010). Adiabat_1ph 3.0 and the MAGMA
        website: educational and research tools for studying the petrology and
        geochemistry of plate margins. AGU Fall Meeting Abstracts 41, ED41B-0644.
"""
from .util.log import Handle

logger = Handle(__name__)

from pyrolite.util.meta import sphinx_doi_link

__doc__ = __doc__.format(
    ghiorso1995=sphinx_doi_link("10.1007/BF00307281"),
    ghiorso2002=sphinx_doi_link("10.1029/2001GC000217"),
    asimow2004=sphinx_doi_link("10.1029/2003GC000568"),
    smith2005=sphinx_doi_link("10.1029/2004GC000816"),
    thompson2007=sphinx_doi_link("10.1093/petrology/egm012"),
)
__doc__ = str(__doc__).replace("ref", __name__)

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

from .download import install_melts

from . import env
from . import automation
from . import vis
from . import tables
